"""
Module with utility functions for reading node attributes.
Contains functions to convert csv files into matrices.

Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
from typing import List, Optional, Tuple
import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split
import random
import os
from .preprocessing import get_names


def read_node_matrix(
        file: str,
        return_coordinates: Optional[bool] = False,
        return_class: Optional[bool] = True,
        remove_prior: Optional[bool] = False,
        remove_morph: Optional[bool] = False,
        ) -> Tuple[np.ndarray, np.ndarray]:
    """
    Read csv and creates X and y matrices.
    Centroids coordinates are removed.
    Labels are subtracted 1 to be in 0-1 range.
    """
    df = pd.read_csv(file)
    remove_vars = ['id', 'class', 'X', 'Y']
    if remove_morph:
        remove_vars.extend([
            'area', 'perimeter', 'std',
            'red0', 'red1', 'red2', 'red3', 'red4',
            'green0', 'green1', 'green2', 'green3', 'green4',
            'blue0', 'blue1', 'blue2', 'blue3', 'blue4'
        ])
    if remove_prior:
        remove_vars.extend(list(filter(lambda x: 'prob' in x, df.columns)))
    if return_class:
        y = df['class'].to_numpy() - 1
        X = df.drop(remove_vars, axis=1).to_numpy()
        if remove_morph and remove_prior:
            X = np.zeros((len(y), 1))
        if not return_coordinates:
            return X, y
        else:
            xx = df['X'].to_numpy()
            yy = df['Y'].to_numpy()
            return X, y, xx, yy
    else:
        remove_vars.remove('class')
        X = df.drop(remove_vars, axis=1).to_numpy()
        if remove_morph and remove_prior:
            X = np.zeros((len(X), 1))
        if not return_coordinates:
            return X, None
        else:
            xx = df['X'].to_numpy()
            yy = df['Y'].to_numpy()
            return X, None, xx, yy


def _read_all_nodes(
        node_dir: str,
        names: List[str],
        remove_prior: Optional[bool] = False,
        remove_morph: Optional[bool] = False,
        ) -> List[np.ndarray]:
    """
    Input
      node_dir: Path to folder with csv files containing node features.
      names: List of files to read. Must have file extension.
    Output
      X: Input data in array format.
      y: Labels in array format.
    """
    X, y = None, None
    for name in names:
        X_, y_ = read_node_matrix(os.path.join(node_dir, name), remove_morph=remove_morph, remove_prior=remove_prior)
        if X is None:
            X = X_  # Shape (n_samples, n_features)
            y = y_  # Shape (n_samples,)
        else:
            X = np.vstack([X, X_])
            y = np.hstack([y, y_])
    return X, y


def read_all_nodes(
        node_dir: str,
        remove_prior: Optional[bool] = False,
        remove_morph: Optional[bool] = False,
        ) -> List[np.ndarray]:
    """
    Input
      node_dir: Path to folder with csv files containing node features.
    Output
      X: Input data in array format.
      y: Labels in array format.
    """
    ext = '.nodes.csv'
    names = get_names(node_dir, ext)
    names = [x + ext for x in names]
    X, y = _read_all_nodes(node_dir, names, remove_morph=remove_morph, remove_prior=remove_prior)
    return X, y


def create_node_splits(
        node_dir: str, val_size: float, test_size: float,
        seed: Optional[int] = None,
        mode: Optional[str] = 'total'
        ) -> List[np.ndarray]:
    """
    Input
      node_dir: Path to folder with csv files containing node features.
      val_size: Percentage of data to use as validation.
      test_size: Percentage of data to use as test.
      seed: Seed for the random split.
      mode: Whether to mix images in the splits or not. It can be 'total' or 'by_img'.
    Output
      X_train, X_val, X_test, y_train, y_val, y_test: Node features and labels.
    """
    ext = '.nodes.csv'
    names = get_names(node_dir, ext)
    names = [x + ext for x in names]
    if mode == 'total':
        X, y = _read_all_nodes(node_dir, names)
        X_tr_val, X_test, y_tr_val, y_test = train_test_split(
            X, y, test_size=test_size, stratify=y, random_state=seed
        )
        X_train, X_val, y_train, y_val = train_test_split(
            X_tr_val, y_tr_val, test_size=val_size / (1 - test_size),
            stratify=y_tr_val, random_state=seed
        )
        return X_train, X_val, X_test, y_train, y_val, y_test
    elif mode == 'by_img':
        random.seed(seed)
        random.shuffle(names)
        N = len(names)
        N_ts = int(N * test_size)
        N_val = int(N * val_size)
        N_tr = N - N_val - N_ts
        train_names = names[:N_tr]
        val_names = names[N_tr:N_val + N_tr]
        test_names = names[N_val + N_tr:]
        X_train, y_train = _read_all_nodes(node_dir, train_names)
        X_val, y_val = _read_all_nodes(node_dir, val_names)
        X_test, y_test = _read_all_nodes(node_dir, test_names)
        return X_train, X_val, X_test, y_train, y_val, y_test
    else:
        assert False, 'Wrong mode.'



"""
Calibration error implementation yet to be merged to sklearn.

Github Thread: https://github.com/scikit-learn/scikit-learn/pull/11096
"""
import numpy as np

from sklearn.utils import assert_all_finite
from sklearn.utils import check_consistent_length
from sklearn.utils import column_or_1d


def calibration_error(y_true, y_prob, sample_weight=None, norm='l2',
                      n_bins=10, strategy='uniform', pos_label=None,
                      reduce_bias=True):
    """Compute calibration error of a binary classifier.
    Across all items in a set of N predictions, the calibration error measures
    the aggregated difference between (1) the average predicted probabilities
    assigned to the positive class, and (2) the frequencies
    of the positive class in the actual outcome.
    The calibration error is only appropriate for binary categorical outcomes.
    Which label is considered to be the positive label is controlled via the
    parameter pos_label, which defaults to 1.
    Parameters
    ----------
    y_true : array-like of shape (n_samples,)
        True targets of a binary classification task.
    y_prob : array-like of (n_samples,)
        Probabilities of the positive class.
    sample_weight : array-like of shape (n_samples,), default=None
        Sample weights.
    norm : {'l1', 'l2', 'max'}, default='l2'
        Norm method. The l1-norm is the Expected Calibration Error (ECE),
        and the max-norm corresponds to Maximum Calibration Error (MCE).
    n_bins : int, default=10
       The number of bins to compute error on.
    strategy : {'uniform', 'quantile'}, default='uniform'
        Strategy used to define the widths of the bins.
        uniform
            All bins have identical widths.
        quantile
            All bins have the same number of points.
    pos_label : int or str, default=None
        Label of the positive class. If None, the maximum label is used as
        positive class.
    reduce_bias : bool, default=True
        Add debiasing term as in Verified Uncertainty Calibration, A. Kumar.
        Only effective for the l2-norm.
    Returns
    -------
    score : float
        calibration error
    Examples
    --------
    >>> import numpy as np
    >>> from sklearn.metrics import calibration_error
    >>> y_true = np.array([0, 0, 0, 1] + [0, 1, 1, 1])
    >>> y_pred = np.array([0.25, 0.25, 0.25, 0.25] + [0.75, 0.75, 0.75, 0.75])
    >>> calibration_error(y_true, y_pred, n_bins=5)
    0.0
    >>> calibration_error(
    ...     y_true, y_pred, n_bins=5, norm="max"
    ... )
    0.0
    >>> y_true = np.array([0, 0, 0, 0] + [1, 1, 1, 1])
    >>> calibration_error(y_true, y_pred, n_bins=5)
    0.25
    >>> calibration_error(
    ...     y_true, y_pred, n_bins=5, norm="max"
    ... )
    0.25
    """
    y_true = column_or_1d(y_true)
    y_prob = column_or_1d(y_prob)
    assert_all_finite(y_true)
    assert_all_finite(y_prob)
    check_consistent_length(y_true, y_prob, sample_weight)
    if any(y_prob < 0) or any(y_prob > 1):
        raise ValueError("y_prob has values outside of [0, 1] range")

    labels = np.unique(y_true)
    if len(labels) > 2:
        raise ValueError("Only binary classification is supported. "
                         "Provided labels %s." % labels)

    if pos_label is None:
        pos_label = y_true.max()
    if pos_label not in labels:
        raise ValueError("pos_label=%r is not a valid label: "
                         "%r" % (pos_label, labels))
    y_true = np.array(y_true == pos_label, int)

    norm_options = ('l1', 'l2', 'max')
    if norm not in norm_options:
        raise ValueError(f'norm has to be one of {norm_options}, got: {norm}.')

    remapping = np.argsort(y_prob)
    y_true = y_true[remapping]
    y_prob = y_prob[remapping]
    if sample_weight is not None:
        sample_weight = sample_weight[remapping]
    else:
        sample_weight = np.ones(y_true.shape[0])

    n_bins = int(n_bins)
    if strategy == 'quantile':
        quantiles = np.percentile(y_prob, np.arange(0, 1, 1.0 / n_bins) * 100)
    elif strategy == 'uniform':
        quantiles = np.arange(0, 1, 1.0 / n_bins)
    else:
        raise ValueError(
            f"Invalid entry to 'strategy' input. Strategy must be either "
            f"'quantile' or 'uniform'. Got {strategy} instead."
        )

    threshold_indices = np.searchsorted(y_prob, quantiles).tolist()
    threshold_indices.append(y_true.shape[0])
    avg_pred_true = np.zeros(n_bins)
    bin_centroid = np.zeros(n_bins)
    delta_count = np.zeros(n_bins)
    debias = np.zeros(n_bins)

    loss = 0.
    count = float(sample_weight.sum())
    for i, i_start in enumerate(threshold_indices[:-1]):
        i_end = threshold_indices[i + 1]
        # ignore empty bins
        if i_end == i_start:
            continue
        delta_count[i] = float(sample_weight[i_start:i_end].sum())
        avg_pred_true[i] = (np.dot(y_true[i_start:i_end],
                                   sample_weight[i_start:i_end])
                            / delta_count[i])
        bin_centroid[i] = (np.dot(y_prob[i_start:i_end],
                                  sample_weight[i_start:i_end])
                           / delta_count[i])
        if norm == "l2" and reduce_bias:
            delta_debias = (
                avg_pred_true[i] * (avg_pred_true[i] - 1) * delta_count[i]
            )
            delta_debias /= (count * delta_count[i] - 1)
            debias[i] = delta_debias

    if norm == "max":
        loss = np.max(np.abs(avg_pred_true - bin_centroid))
    elif norm == "l1":
        delta_loss = np.abs(avg_pred_true - bin_centroid) * delta_count
        loss = np.sum(delta_loss) / count
    elif norm == "l2":
        delta_loss = (avg_pred_true - bin_centroid)**2 * delta_count
        loss = np.sum(delta_loss) / count
        if reduce_bias:
            loss += np.sum(debias)
        loss = np.sqrt(max(loss, 0.))
    return loss



import numpy as np
import matplotlib.pyplot as plt
import torch
import torch.nn.functional as F

from ...misc.utils import cropping_center
from .utils import dice_loss, mse_loss, msge_loss, xentropy_loss

from collections import OrderedDict
from typing import Tuple

####
def train_step(batch_data, run_info):
    # TODO: synchronize the attach protocol
    run_info, state_info = run_info
    loss_func_dict = {
        "bce": xentropy_loss,
        "dice": dice_loss,
        "mse": mse_loss,
        "msge": msge_loss,
    }
    # use 'ema' to add for EMA calculation, must be scalar!
    result_dict = {"EMA": {}}
    track_value = lambda name, value: result_dict["EMA"].update({name: value})

    ####
    model = run_info["net"]["desc"]
    optimizer = run_info["net"]["optimizer"]

    ####
    imgs = batch_data["img"]
    true_np = batch_data["np_map"]
    true_hv = batch_data["hv_map"]
    if run_info["net"]["use_cpu"]:
        imgs = imgs.type(torch.float32)  # to NCHW    
    else:
        imgs = imgs.to("cuda").type(torch.float32)  # to NCHW
    imgs = imgs.permute(0, 3, 1, 2).contiguous()

    # HWC
    if run_info["net"]["use_cpu"]:
        true_np = true_np.type(torch.int64)
        true_hv = true_hv.type(torch.float32)
    else:
        true_np = true_np.to("cuda").type(torch.int64)
        true_hv = true_hv.to("cuda").type(torch.float32)

    true_np_onehot = (F.one_hot(true_np, num_classes=2)).type(torch.float32)
    true_dict = {
        "np": true_np_onehot,
        "hv": true_hv,
    }
    
    if model.module.nr_types is not None:
        true_tp = batch_data["tp_map"]
        if run_info["net"]["use_cpu"]:
            true_tp = torch.squeeze(true_tp).type(torch.int64)
        else:
            true_tp = torch.squeeze(true_tp).to("cuda").type(torch.int64)
        true_tp_onehot = F.one_hot(true_tp, num_classes=model.module.nr_types)
        true_tp_onehot = true_tp_onehot.type(torch.float32)
        true_dict["tp"] = true_tp_onehot

    ####
    model.train()
    model.zero_grad()  # not rnn so not accumulate

    pred_dict = model(imgs)
    pred_dict = OrderedDict(
        [[k, v.permute(0, 2, 3, 1).contiguous()] for k, v in pred_dict.items()]
    )
    pred_dict["np"] = F.softmax(pred_dict["np"], dim=-1)
    if model.module.nr_types is not None:
        pred_dict["tp"] = F.softmax(pred_dict["tp"], dim=-1)

    ####
    loss = 0
    loss_opts = run_info["net"]["extra_info"]["loss"]
    for branch_name in pred_dict.keys():
        for loss_name, loss_weight in loss_opts[branch_name].items():
            loss_func = loss_func_dict[loss_name]
            loss_args = [true_dict[branch_name], pred_dict[branch_name]]
            if loss_name == "msge":
                loss_args.append(true_np_onehot[..., 1])
            term_loss = loss_func(*loss_args, device="cpu" if run_info["net"]["use_cpu"] else "cuda")
            track_value("loss_%s_%s" % (branch_name, loss_name), term_loss.cpu().item())
            loss += loss_weight * term_loss

    track_value("overall_loss", loss.cpu().item())
    # * gradient update

    # torch.set_printoptions(precision=10)
    loss.backward()
    optimizer.step()
    ####

    # pick 2 random sample from the batch for visualization
    sample_indices = torch.randint(0, true_np.shape[0], (2,))

    imgs = (imgs[sample_indices]).byte()  # to uint8
    imgs = imgs.permute(0, 2, 3, 1).contiguous().cpu().numpy()

    pred_dict["np"] = pred_dict["np"][..., 1]  # return pos only
    pred_dict = {
        k: v[sample_indices].detach().cpu().numpy() for k, v in pred_dict.items()
    }

    true_dict["np"] = true_np
    true_dict = {
        k: v[sample_indices].detach().cpu().numpy() for k, v in true_dict.items()
    }

    # * Its up to user to define the protocol to process the raw output per step!
    result_dict["raw"] = {  # protocol for contents exchange within `raw`
        "img": imgs,
        "np": (true_dict["np"], pred_dict["np"]),
        "hv": (true_dict["hv"], pred_dict["hv"]),
    }
    return result_dict


####
def valid_step(batch_data, run_info):
    run_info, state_info = run_info
    ####
    model = run_info["net"]["desc"]
    model.eval()  # infer mode

    ####
    imgs = batch_data["img"]
    true_np = batch_data["np_map"]
    true_hv = batch_data["hv_map"]

    if run_info["net"]["use_cpu"]:
        imgs_gpu = imgs.type(torch.float32)  # to NCHW
    else:
        imgs_gpu = imgs.to("cuda").type(torch.float32)  # to NCHW
    imgs_gpu = imgs_gpu.permute(0, 3, 1, 2).contiguous()

    # HWC
    true_np = torch.squeeze(true_np).type(torch.int64)
    true_hv = torch.squeeze(true_hv).type(torch.float32)

    true_dict = {
        "np": true_np,
        "hv": true_hv,
    }

    if model.module.nr_types is not None:
        true_tp = batch_data["tp_map"]
        true_tp = torch.squeeze(true_tp).type(torch.int64)
        true_dict["tp"] = true_tp

    # --------------------------------------------------------------
    with torch.no_grad():  # dont compute gradient
        pred_dict = model(imgs_gpu)
        pred_dict = OrderedDict(
            [[k, v.permute(0, 2, 3, 1).contiguous()] for k, v in pred_dict.items()]
        )
        pred_dict["np"] = F.softmax(pred_dict["np"], dim=-1)[..., 1]
        if model.module.nr_types is not None:
            type_map = F.softmax(pred_dict["tp"], dim=-1)
            type_map = torch.argmax(type_map, dim=-1, keepdim=False)
            type_map = type_map.type(torch.float32)
            pred_dict["tp"] = type_map

    # * Its up to user to define the protocol to process the raw output per step!
    result_dict = {  # protocol for contents exchange within `raw`
        "raw": {
            "imgs": imgs.numpy(),
            "true_np": true_dict["np"].numpy(),
            "true_hv": true_dict["hv"].numpy(),
            "prob_np": pred_dict["np"].cpu().numpy(),
            "pred_hv": pred_dict["hv"].cpu().numpy(),
        }
    }
    if model.module.nr_types is not None:
        result_dict["raw"]["true_tp"] = true_dict["tp"].numpy()
        result_dict["raw"]["pred_tp"] = pred_dict["tp"].cpu().numpy()
    return result_dict


####
def infer_step(batch_data, model, use_cpu=False):

    ####
    patch_imgs = batch_data

    if use_cpu:
        patch_imgs_gpu = patch_imgs.type(torch.float32)  # to NCHW
    else:
        patch_imgs_gpu = patch_imgs.to("cuda").type(torch.float32)  # to NCHW
    patch_imgs_gpu = patch_imgs_gpu.permute(0, 3, 1, 2).contiguous()

    ####
    model.eval()  # infer mode

    # --------------------------------------------------------------
    with torch.no_grad():  # dont compute gradient
        pred_dict = model(patch_imgs_gpu)
        pred_dict = OrderedDict(
            [[k, v.permute(0, 2, 3, 1).contiguous()] for k, v in pred_dict.items()]
        )
        pred_dict["np"] = F.softmax(pred_dict["np"], dim=-1)[..., 1:]
        if "tp" in pred_dict:
            type_map = F.softmax(pred_dict["tp"], dim=-1)
            type_map = torch.argmax(type_map, dim=-1, keepdim=True)
            type_map = type_map.type(torch.float32)
            pred_dict["tp"] = type_map
        pred_output = torch.cat(list(pred_dict.values()), -1)

    # * Its up to user to define the protocol to process the raw output per step!
    return pred_output.cpu().numpy()


def fix_shape(x: Tuple) -> Tuple:
    if len(x) != 3:
        return x
    return (x[0], x[1], x[2], 1)

####
def viz_step_output(raw_data, nr_types=None):
    """
    `raw_data` will be implicitly provided in the similar format as the 
    return dict from train/valid step, but may have been accumulated across N running step
    """
    imgs = raw_data["img"]
    true_np, pred_np = raw_data["np"]
    true_np_shape, pred_np_shape = fix_shape(true_np.shape), fix_shape(pred_np.shape) 
    true_hv, pred_hv = raw_data["hv"]
    if nr_types is not None:
        true_tp, pred_tp = raw_data["tp"]

    aligned_shape = [list(imgs.shape), list(true_np_shape), list(pred_np_shape)]
    # aligned_shape = [list(imgs.shape), list(pred_np.shape)]
    aligned_shape = np.min(np.array(aligned_shape), axis=0)[1:3]

    cmap = plt.get_cmap("jet")

    def colorize(ch, vmin, vmax):
        """
        Will clamp value value outside the provided range to vmax and vmin
        """
        ch = np.squeeze(ch.astype("float32"))
        ch[ch > vmax] = vmax  # clamp value
        ch[ch < vmin] = vmin
        ch = (ch - vmin) / (vmax - vmin + 1.0e-16)
        # take RGB from RGBA heat map
        ch_cmap = (cmap(ch)[..., :3] * 255).astype("uint8")
        # ch_cmap = center_pad_to_shape(ch_cmap, aligned_shape)
        return ch_cmap

    viz_list = []
    for idx in range(imgs.shape[0]):
        # img = center_pad_to_shape(imgs[idx], aligned_shape)
        img = cropping_center(imgs[idx], aligned_shape)

        true_viz_list = [img]
        # cmap may randomly fails if of other types
        true_viz_list.append(colorize(true_np[idx], 0, 1))
        true_viz_list.append(colorize(true_hv[idx][..., 0], -1, 1))
        true_viz_list.append(colorize(true_hv[idx][..., 1], -1, 1))
        if nr_types is not None:  # TODO: a way to pass through external info
            true_viz_list.append(colorize(true_tp[idx], 0, nr_types))
        true_viz_list = np.concatenate(true_viz_list, axis=1)

        pred_viz_list = [img]
        # cmap may randomly fails if of other types
        pred_viz_list.append(colorize(pred_np[idx], 0, 1))
        pred_viz_list.append(colorize(pred_hv[idx][..., 0], -1, 1))
        pred_viz_list.append(colorize(pred_hv[idx][..., 1], -1, 1))
        if nr_types is not None:
            pred_viz_list.append(colorize(pred_tp[idx], 0, nr_types))
        pred_viz_list = np.concatenate(pred_viz_list, axis=1)

        viz_list.append(np.concatenate([true_viz_list, pred_viz_list], axis=0))
    viz_list = np.concatenate(viz_list, axis=0)
    return viz_list


####
from itertools import chain


def proc_valid_step_output(raw_data, nr_types=None):
    # TODO: add auto populate from main state track list
    track_dict = {"scalar": {}, "image": {}}

    def track_value(name, value, vtype):
        return track_dict[vtype].update({name: value})

    def _dice_info(true, pred, label):
        true = np.array(true == label, np.int32)
        pred = np.array(pred == label, np.int32)
        inter = (pred * true).sum()
        total = (pred + true).sum()
        return inter, total

    over_inter = 0
    over_total = 0
    over_correct = 0
    prob_np = raw_data["prob_np"]
    true_np = raw_data["true_np"]
    for idx in range(len(raw_data["true_np"])):
    # for idx in range(len(raw_data["prob_np"])):
        patch_prob_np = prob_np[idx]
        patch_true_np = true_np[idx]
        patch_pred_np = np.array(patch_prob_np > 0.5, dtype=np.int32)
        inter, total = _dice_info(patch_true_np, patch_pred_np, 1)
        correct = (patch_pred_np == patch_true_np).sum()
        over_inter += inter
        over_total += total
        over_correct += correct
    nr_pixels = len(true_np) * np.size(true_np[0])
    acc_np = over_correct / nr_pixels
    dice_np = 2 * over_inter / (over_total + 1.0e-8)
    track_value("np_acc", acc_np, "scalar")
    track_value("np_dice", dice_np, "scalar")

    # * TP statistic
    if nr_types is not None:
        pred_tp = raw_data["pred_tp"]
        true_tp = raw_data["true_tp"]
        for type_id in range(0, nr_types):
            over_inter = 0
            over_total = 0
            for idx in range(len(raw_data["true_np"])):
            # for idx in range(len(raw_data["pred_tp"])):
                patch_pred_tp = pred_tp[idx]
                patch_true_tp = true_tp[idx]
                inter, total = _dice_info(patch_true_tp, patch_pred_tp, type_id)
                over_inter += inter
                over_total += total
            dice_tp = 2 * over_inter / (over_total + 1.0e-8)
            track_value("tp_dice_%d" % type_id, dice_tp, "scalar")

    # * HV regression statistic
    pred_hv = raw_data["pred_hv"]
    true_hv = raw_data["true_hv"]

    over_squared_error = 0
    for idx in range(len(raw_data["true_np"])):
    # for idx in range(len(raw_data["pred_hv"])):
        patch_pred_hv = pred_hv[idx]
        patch_true_hv = true_hv[idx]
        squared_error = patch_pred_hv - patch_true_hv
        squared_error = squared_error * squared_error
        over_squared_error += squared_error.sum()
    mse = over_squared_error / nr_pixels
    track_value("hv_mse", mse, "scalar")

    # *
    imgs = raw_data["imgs"]
    selected_idx = np.random.randint(0, len(imgs), size=(8,)).tolist()
    imgs = np.array([imgs[idx] for idx in selected_idx])
    true_np = np.array([true_np[idx] for idx in selected_idx])
    true_hv = np.array([true_hv[idx] for idx in selected_idx])
    prob_np = np.array([prob_np[idx] for idx in selected_idx])
    pred_hv = np.array([pred_hv[idx] for idx in selected_idx])
    viz_raw_data = {"img": imgs, "np": (true_np, prob_np), "hv": (true_hv, pred_hv)}

    if nr_types is not None:
        true_tp = np.array([true_tp[idx] for idx in selected_idx])
        pred_tp = np.array([pred_tp[idx] for idx in selected_idx])
        viz_raw_data["tp"] = (true_tp, pred_tp)
    viz_fig = viz_step_output(viz_raw_data, nr_types)
    track_dict["image"]["output"] = viz_fig

    return track_dict



import numpy as np
import scipy
from scipy.optimize import linear_sum_assignment


# --------------------------Optimised for Speed
def get_fast_aji(true, pred):
    """AJI version distributed by MoNuSeg, has no permutation problem but suffered from 
    over-penalisation similar to DICE2.

    Fast computation requires instance IDs are in contiguous orderding i.e [1, 2, 3, 4] 
    not [2, 3, 6, 10]. Please call `remap_label` before hand and `by_size` flag has no 
    effect on the result.

    """
    true = np.copy(true)  # ? do we need this
    pred = np.copy(pred)
    true_id_list = list(np.unique(true))
    pred_id_list = list(np.unique(pred))

    true_masks = [
        None,
    ]
    for t in true_id_list[1:]:
        t_mask = np.array(true == t, np.uint8)
        true_masks.append(t_mask)

    pred_masks = [
        None,
    ]
    for p in pred_id_list[1:]:
        p_mask = np.array(pred == p, np.uint8)
        pred_masks.append(p_mask)

    # prefill with value
    pairwise_inter = np.zeros(
        [len(true_id_list) - 1, len(pred_id_list) - 1], dtype=np.float64
    )
    pairwise_union = np.zeros(
        [len(true_id_list) - 1, len(pred_id_list) - 1], dtype=np.float64
    )

    # caching pairwise
    for true_id in true_id_list[1:]:  # 0-th is background
        t_mask = true_masks[true_id]
        pred_true_overlap = pred[t_mask > 0]
        pred_true_overlap_id = np.unique(pred_true_overlap)
        pred_true_overlap_id = list(pred_true_overlap_id)
        for pred_id in pred_true_overlap_id:
            if pred_id == 0:  # ignore
                continue  # overlaping background
            p_mask = pred_masks[pred_id]
            total = (t_mask + p_mask).sum()
            inter = (t_mask * p_mask).sum()
            pairwise_inter[true_id - 1, pred_id - 1] = inter
            pairwise_union[true_id - 1, pred_id - 1] = total - inter

    pairwise_iou = pairwise_inter / (pairwise_union + 1.0e-6)
    # pair of pred that give highest iou for each true, dont care
    # about reusing pred instance multiple times
    paired_pred = np.argmax(pairwise_iou, axis=1)
    pairwise_iou = np.max(pairwise_iou, axis=1)
    # exlude those dont have intersection
    paired_true = np.nonzero(pairwise_iou > 0.0)[0]
    paired_pred = paired_pred[paired_true]
    # print(paired_true.shape, paired_pred.shape)
    overall_inter = (pairwise_inter[paired_true, paired_pred]).sum()
    overall_union = (pairwise_union[paired_true, paired_pred]).sum()

    paired_true = list(paired_true + 1)  # index to instance ID
    paired_pred = list(paired_pred + 1)
    # add all unpaired GT and Prediction into the union
    unpaired_true = np.array(
        [idx for idx in true_id_list[1:] if idx not in paired_true]
    )
    unpaired_pred = np.array(
        [idx for idx in pred_id_list[1:] if idx not in paired_pred]
    )
    for true_id in unpaired_true:
        overall_union += true_masks[true_id].sum()
    for pred_id in unpaired_pred:
        overall_union += pred_masks[pred_id].sum()

    aji_score = overall_inter / overall_union
    return aji_score


#####
def get_fast_aji_plus(true, pred):
    """AJI+, an AJI version with maximal unique pairing to obtain overall intersecion.
    Every prediction instance is paired with at most 1 GT instance (1 to 1) mapping, unlike AJI 
    where a prediction instance can be paired against many GT instances (1 to many).
    Remaining unpaired GT and Prediction instances will be added to the overall union.
    The 1 to 1 mapping prevents AJI's over-penalisation from happening.

    Fast computation requires instance IDs are in contiguous orderding i.e [1, 2, 3, 4] 
    not [2, 3, 6, 10]. Please call `remap_label` before hand and `by_size` flag has no 
    effect on the result.

    """
    true = np.copy(true)  # ? do we need this
    pred = np.copy(pred)
    true_id_list = list(np.unique(true))
    pred_id_list = list(np.unique(pred))

    true_masks = [
        None,
    ]
    for t in true_id_list[1:]:
        t_mask = np.array(true == t, np.uint8)
        true_masks.append(t_mask)

    pred_masks = [
        None,
    ]
    for p in pred_id_list[1:]:
        p_mask = np.array(pred == p, np.uint8)
        pred_masks.append(p_mask)

    # prefill with value
    pairwise_inter = np.zeros(
        [len(true_id_list) - 1, len(pred_id_list) - 1], dtype=np.float64
    )
    pairwise_union = np.zeros(
        [len(true_id_list) - 1, len(pred_id_list) - 1], dtype=np.float64
    )

    # caching pairwise
    for true_id in true_id_list[1:]:  # 0-th is background
        t_mask = true_masks[true_id]
        pred_true_overlap = pred[t_mask > 0]
        pred_true_overlap_id = np.unique(pred_true_overlap)
        pred_true_overlap_id = list(pred_true_overlap_id)
        for pred_id in pred_true_overlap_id:
            if pred_id == 0:  # ignore
                continue  # overlaping background
            p_mask = pred_masks[pred_id]
            total = (t_mask + p_mask).sum()
            inter = (t_mask * p_mask).sum()
            pairwise_inter[true_id - 1, pred_id - 1] = inter
            pairwise_union[true_id - 1, pred_id - 1] = total - inter
    #
    pairwise_iou = pairwise_inter / (pairwise_union + 1.0e-6)
    #### Munkres pairing to find maximal unique pairing
    paired_true, paired_pred = linear_sum_assignment(-pairwise_iou)
    ### extract the paired cost and remove invalid pair
    paired_iou = pairwise_iou[paired_true, paired_pred]
    # now select all those paired with iou != 0.0 i.e have intersection
    paired_true = paired_true[paired_iou > 0.0]
    paired_pred = paired_pred[paired_iou > 0.0]
    paired_inter = pairwise_inter[paired_true, paired_pred]
    paired_union = pairwise_union[paired_true, paired_pred]
    paired_true = list(paired_true + 1)  # index to instance ID
    paired_pred = list(paired_pred + 1)
    overall_inter = paired_inter.sum()
    overall_union = paired_union.sum()
    # add all unpaired GT and Prediction into the union
    unpaired_true = np.array(
        [idx for idx in true_id_list[1:] if idx not in paired_true]
    )
    unpaired_pred = np.array(
        [idx for idx in pred_id_list[1:] if idx not in paired_pred]
    )
    for true_id in unpaired_true:
        overall_union += true_masks[true_id].sum()
    for pred_id in unpaired_pred:
        overall_union += pred_masks[pred_id].sum()
    #
    aji_score = overall_inter / overall_union
    return aji_score


#####
def get_fast_pq(true, pred, match_iou=0.5):
    """`match_iou` is the IoU threshold level to determine the pairing between
    GT instances `p` and prediction instances `g`. `p` and `g` is a pair
    if IoU > `match_iou`. However, pair of `p` and `g` must be unique 
    (1 prediction instance to 1 GT instance mapping).

    If `match_iou` < 0.5, Munkres assignment (solving minimum weight matching
    in bipartite graphs) is caculated to find the maximal amount of unique pairing. 

    If `match_iou` >= 0.5, all IoU(p,g) > 0.5 pairing is proven to be unique and
    the number of pairs is also maximal.    
    
    Fast computation requires instance IDs are in contiguous orderding 
    i.e [1, 2, 3, 4] not [2, 3, 6, 10]. Please call `remap_label` beforehand 
    and `by_size` flag has no effect on the result.

    Returns:
        [dq, sq, pq]: measurement statistic

        [paired_true, paired_pred, unpaired_true, unpaired_pred]: 
                      pairing information to perform measurement
                    
    """
    assert match_iou >= 0.0, "Cant' be negative"

    true = np.copy(true)
    pred = np.copy(pred)
    true_id_list = list(np.unique(true))
    pred_id_list = list(np.unique(pred))

    true_masks = [
        None,
    ]
    for t in true_id_list[1:]:
        t_mask = np.array(true == t, np.uint8)
        true_masks.append(t_mask)

    pred_masks = [
        None,
    ]
    for p in pred_id_list[1:]:
        p_mask = np.array(pred == p, np.uint8)
        pred_masks.append(p_mask)

    # prefill with value
    pairwise_iou = np.zeros(
        [len(true_id_list) - 1, len(pred_id_list) - 1], dtype=np.float64
    )

    # caching pairwise iou
    for true_id in true_id_list[1:]:  # 0-th is background
        t_mask = true_masks[true_id]
        pred_true_overlap = pred[t_mask > 0]
        pred_true_overlap_id = np.unique(pred_true_overlap)
        pred_true_overlap_id = list(pred_true_overlap_id)
        for pred_id in pred_true_overlap_id:
            if pred_id == 0:  # ignore
                continue  # overlaping background
            p_mask = pred_masks[pred_id]
            total = (t_mask + p_mask).sum()
            inter = (t_mask * p_mask).sum()
            iou = inter / (total - inter)
            pairwise_iou[true_id - 1, pred_id - 1] = iou
    #
    if match_iou >= 0.5:
        paired_iou = pairwise_iou[pairwise_iou > match_iou]
        pairwise_iou[pairwise_iou <= match_iou] = 0.0
        paired_true, paired_pred = np.nonzero(pairwise_iou)
        paired_iou = pairwise_iou[paired_true, paired_pred]
        paired_true += 1  # index is instance id - 1
        paired_pred += 1  # hence return back to original
    else:  # * Exhaustive maximal unique pairing
        #### Munkres pairing with scipy library
        # the algorithm return (row indices, matched column indices)
        # if there is multiple same cost in a row, index of first occurence
        # is return, thus the unique pairing is ensure
        # inverse pair to get high IoU as minimum
        paired_true, paired_pred = linear_sum_assignment(-pairwise_iou)
        ### extract the paired cost and remove invalid pair
        paired_iou = pairwise_iou[paired_true, paired_pred]

        # now select those above threshold level
        # paired with iou = 0.0 i.e no intersection => FP or FN
        paired_true = list(paired_true[paired_iou > match_iou] + 1)
        paired_pred = list(paired_pred[paired_iou > match_iou] + 1)
        paired_iou = paired_iou[paired_iou > match_iou]

    # get the actual FP and FN
    unpaired_true = [idx for idx in true_id_list[1:] if idx not in paired_true]
    unpaired_pred = [idx for idx in pred_id_list[1:] if idx not in paired_pred]
    # print(paired_iou.shape, paired_true.shape, len(unpaired_true), len(unpaired_pred))

    #
    tp = len(paired_true)
    fp = len(unpaired_pred)
    fn = len(unpaired_true)
    # get the F1-score i.e DQ
    dq = tp / (tp + 0.5 * fp + 0.5 * fn)
    # get the SQ, no paired has 0 iou so not impact
    sq = paired_iou.sum() / (tp + 1.0e-6)

    return [dq, sq, dq * sq], [paired_true, paired_pred, unpaired_true, unpaired_pred]


#####
def get_fast_dice_2(true, pred):
    """Ensemble dice."""
    true = np.copy(true)
    pred = np.copy(pred)
    true_id = list(np.unique(true))
    pred_id = list(np.unique(pred))

    overall_total = 0
    overall_inter = 0

    true_masks = [np.zeros(true.shape)]
    for t in true_id[1:]:
        t_mask = np.array(true == t, np.uint8)
        true_masks.append(t_mask)

    pred_masks = [np.zeros(true.shape)]
    for p in pred_id[1:]:
        p_mask = np.array(pred == p, np.uint8)
        pred_masks.append(p_mask)

    for true_idx in range(1, len(true_id)):
        t_mask = true_masks[true_idx]
        pred_true_overlap = pred[t_mask > 0]
        pred_true_overlap_id = np.unique(pred_true_overlap)
        pred_true_overlap_id = list(pred_true_overlap_id)
        try:  # blinly remove background
            pred_true_overlap_id.remove(0)
        except ValueError:
            pass  # just mean no background
        for pred_idx in pred_true_overlap_id:
            p_mask = pred_masks[pred_idx]
            total = (t_mask + p_mask).sum()
            inter = (t_mask * p_mask).sum()
            overall_total += total
            overall_inter += inter

    return 2 * overall_inter / overall_total


#####--------------------------As pseudocode
def get_dice_1(true, pred):
    """Traditional dice."""
    # cast to binary 1st
    true = np.copy(true)
    pred = np.copy(pred)
    true[true > 0] = 1
    pred[pred > 0] = 1
    inter = true * pred
    denom = true + pred
    return 2.0 * np.sum(inter) / np.sum(denom)


####
def get_dice_2(true, pred):
    """Ensemble Dice as used in Computational Precision Medicine Challenge."""
    true = np.copy(true)
    pred = np.copy(pred)
    true_id = list(np.unique(true))
    pred_id = list(np.unique(pred))
    # remove background aka id 0
    true_id.remove(0)
    pred_id.remove(0)

    total_markup = 0
    total_intersect = 0
    for t in true_id:
        t_mask = np.array(true == t, np.uint8)
        for p in pred_id:
            p_mask = np.array(pred == p, np.uint8)
            intersect = p_mask * t_mask
            if intersect.sum() > 0:
                total_intersect += intersect.sum()
                total_markup += t_mask.sum() + p_mask.sum()
    return 2 * total_intersect / total_markup


#####
def remap_label(pred, by_size=False):
    """Rename all instance id so that the id is contiguous i.e [0, 1, 2, 3] 
    not [0, 2, 4, 6]. The ordering of instances (which one comes first) 
    is preserved unless by_size=True, then the instances will be reordered
    so that bigger nucler has smaller ID.

    Args:
        pred    : the 2d array contain instances where each instances is marked
                  by non-zero integer
        by_size : renaming with larger nuclei has smaller id (on-top)

    """
    pred_id = list(np.unique(pred))
    pred_id.remove(0)
    if len(pred_id) == 0:
        return pred  # no label
    if by_size:
        pred_size = []
        for inst_id in pred_id:
            size = (pred == inst_id).sum()
            pred_size.append(size)
        # sort the id by size in descending order
        pair_list = zip(pred_id, pred_size)
        pair_list = sorted(pair_list, key=lambda x: x[1], reverse=True)
        pred_id, pred_size = zip(*pair_list)

    new_pred = np.zeros(pred.shape, np.int32)
    for idx, inst_id in enumerate(pred_id):
        new_pred[pred == inst_id] = idx + 1
    return new_pred


#####
def pair_coordinates(setA, setB, radius):
    """Use the Munkres or Kuhn-Munkres algorithm to find the most optimal 
    unique pairing (largest possible match) when pairing points in set B 
    against points in set A, using distance as cost function.

    Args:
        setA, setB: np.array (float32) of size Nx2 contains the of XY coordinate
                    of N different points 
        radius: valid area around a point in setA to consider 
                a given coordinate in setB a candidate for match
    Return:
        pairing: pairing is an array of indices
        where point at index pairing[0] in set A paired with point
        in set B at index pairing[1]
        unparedA, unpairedB: remaining poitn in set A and set B unpaired

    """
    # * Euclidean distance as the cost matrix
    pair_distance = scipy.spatial.distance.cdist(setA, setB, metric='euclidean')

    # * Munkres pairing with scipy library
    # the algorithm return (row indices, matched column indices)
    # if there is multiple same cost in a row, index of first occurence 
    # is return, thus the unique pairing is ensured
    indicesA, paired_indicesB = linear_sum_assignment(pair_distance)

    # extract the paired cost and remove instances 
    # outside of designated radius
    pair_cost = pair_distance[indicesA, paired_indicesB]

    pairedA = indicesA[pair_cost <= radius]
    pairedB = paired_indicesB[pair_cost <= radius]

    pairing = np.concatenate([pairedA[:,None], pairedB[:,None]], axis=-1)
    unpairedA = np.delete(np.arange(setA.shape[0]), pairedA)
    unpairedB = np.delete(np.arange(setB.shape[0]), pairedB)
    return pairing, unpairedA, unpairedB



"""
Computes a graph representation from the images and pngcsv labels.

Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import argparse
import pandas as pd
import numpy as np
import os
from concurrent.futures import ThreadPoolExecutor
import logging
from tqdm import tqdm
from ..utils.preprocessing import get_mask, read_labels, parse_path, create_dir, save_graph, get_names, apply_mask, extract_features, add_node, read_image


def pngcsv2graph(img: np.ndarray, png: np.ndarray, csv: pd.DataFrame) -> pd.DataFrame:
    """
    Converts an original image and PNGCSV labels into a graph representation.

    Given the original image and PNGCSV labels, this function extracts attributes
    from each label and constructs a graph representation in the form of a DataFrame.

    The current attributes extracted for each label include:
        - X, Y coordinates of the centroid
        - Area
        - Perimeter
        - Variance
        - Regularity (not yet implemented)
        - Histogram (5 bins)

    :param img: The original image.
    :type img: np.ndarray
    :param png: The PNGCSV labels.
    :type png: np.ndarray
    :param csv: The CSV file containing label information.
    :type csv: pd.DataFrame
    :return: The graph representation in the form of a DataFrame.
    :rtype: pd.DataFrame
    """
    graph = {}
    for idx, cls in csv.itertuples(index=False, name=None):
        mask = get_mask(png, idx)
        msk_img, msk, X, Y = apply_mask(img, mask)
        try:
            feats = extract_features(msk_img, msk)
        except Exception:
            feats = extract_features(msk_img, msk, debug=True)
        if len(feats) > 0:
            feats['class'] = cls
            feats['id'] = idx
            feats['X'] = X
            feats['Y'] = Y
        add_node(graph, feats)
    return pd.DataFrame(graph)


def _create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--png-dir', type=str, required=True,
                        help='Path to png files.')
    parser.add_argument('--csv-dir', type=str, required=True,
                        help='Path to csv files.')
    parser.add_argument('--orig-dir', type=str, required=True,
                        help='Path to original images.')
    parser.add_argument('--output-path', type=str, required=True,
                        help='Path to save files.')
    parser.add_argument('--num-workers', type=int, default=1)
    return parser


def main_subthread(
        name: str,
        png_dir: str,
        csv_dir: str,
        orig_dir: str,
        output_path: str,
        pbar: tqdm,
        ) -> None:
    """
    Wrapper to use multiprocessing
    """
    try:
        png, csv = read_labels(name, png_dir, csv_dir)
        img = read_image(name, orig_dir)
        graph = pngcsv2graph(img, png, csv)
        save_graph(graph, os.path.join(output_path, name + '.nodes.csv'))
    except Exception as e:
        logging.warning(e)
        logging.warning('Failed at:', name)
    finally:
        pbar.update(1)


def main_with_args(args):
    png_dir = parse_path(args.png_dir)
    csv_dir = parse_path(args.csv_dir)
    orig_dir = parse_path(args.orig_dir)
    output_path = parse_path(args.output_path)
    create_dir(output_path)

    names = get_names(png_dir, '.GT_cells.png')
    pbar = tqdm(total=len(names))
    if args.num_workers > 0:
        with ThreadPoolExecutor(max_workers=args.num_workers) as executor:
            for name in names:
                executor.submit(main_subthread, name, png_dir, csv_dir, orig_dir, output_path, pbar)
    else:
        for name in names:
            main_subthread(name, png_dir, csv_dir, orig_dir, output_path, pbar)


def main():
    parser = _create_parser()
    args = parser.parse_args()
    main_with_args(args)



"""
Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import pytest
import os
import json

from tumourkit.postprocessing.rswoosh import rswoosh
from tumourkit.utils.preprocessing import get_names, parse_path
from tumourkit.utils.postprocessing import create_comparator, merge_cells

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
RSWOOSH_DIR = parse_path(TEST_DIR) + 'rswoosh/'
THRESHOLD = 1
NUM_FRONTIER = 30

@pytest.mark.parametrize("name", get_names(RSWOOSH_DIR, '.input.json'))
def test_conversion(name):
    with open(RSWOOSH_DIR + name + '.input.json', 'r') as f:
        inp = json.load(f)
    with open(RSWOOSH_DIR + name + '.output.json', 'r') as f:
        expected_out = json.load(f)
    predicted_out = rswoosh(inp, create_comparator(THRESHOLD, NUM_FRONTIER), merge_cells)
    if not len(predicted_out) == len(expected_out):
        print(name)
        assert(False)
    assert(True)



"""run_infer.py

Usage:
  run_infer.py [options] [--help] <command> [<args>...]
  run_infer.py --version
  run_infer.py (-h | --help)

Options:
  -h --help                   Show this string.
  --version                   Show version.

  --gpu=<id>                  GPU list. [default: 0]
  --nr_types=<n>              Number of nuclei types to predict. [default: 0]
  --type_info_path=<path>     Path to a json define mapping between type id, type name, 
                              and expected overlaid color. [default: '']

  --model_path=<path>         Path to saved checkpoint.
  --model_mode=<mode>         Original HoVer-Net or the reduced version used PanNuke and MoNuSAC, 
                              'original' or 'fast'. [default: fast]
  --shape=<shape>             Either 270 or 518.
  --nr_inference_workers=<n>  Number of workers during inference. [default: 8]
  --nr_post_proc_workers=<n>  Number of workers during post-processing. [default: 16]
  --batch_size=<n>            Batch size per 1 GPU. [default: 32]

Two command mode are `tile` and `wsi` to enter corresponding inference mode
    tile  run the inference on tile
    wsi   run the inference on wsi

Use `run_infer.py <command> --help` to show their options and usage.
"""

tile_cli = """
Arguments for processing tiles.

usage:
    tile (--input_dir=<path>) (--output_dir=<path>) \
         [--draw_dot] [--save_qupath] [--save_raw_map] [--mem_usage=<n>]
    
options:
   --input_dir=<path>     Path to input data directory. Assumes the files are not nested within directory.
   --output_dir=<path>    Path to output directory..

   --mem_usage=<n>        Declare how much memory (physical + swap) should be used for caching. 
                          By default it will load as many tiles as possible till reaching the 
                          declared limit. [default: 0.2]
   --draw_dot             To draw nuclei centroid on overlay. [default: False]
   --save_qupath          To optionally output QuPath v0.2.3 compatible format. [default: False]
   --save_raw_map         To save raw prediction or not. [default: False]
"""

wsi_cli = """
Arguments for processing wsi

usage:
    wsi (--input_dir=<path>) (--output_dir=<path>) [--proc_mag=<n>]\
        [--cache_path=<path>] [--input_mask_dir=<path>] \
        [--ambiguous_size=<n>] [--chunk_shape=<n>] [--tile_shape=<n>] \
        [--save_thumb] [--save_mask]
    
options:
    --input_dir=<path>      Path to input data directory. Assumes the files are not nested within directory.
    --output_dir=<path>     Path to output directory.
    --cache_path=<path>     Path for cache. Should be placed on SSD with at least 100GB. [default: cache]
    --mask_dir=<path>       Path to directory containing tissue masks. 
                            Should have the same name as corresponding WSIs. [default: '']

    --proc_mag=<n>          Magnification level (objective power) used for WSI processing. [default: 40]
    --ambiguous_size=<int>  Define ambiguous region along tiling grid to perform re-post processing. [default: 128]
    --chunk_shape=<n>       Shape of chunk for processing. [default: 10000]
    --tile_shape=<n>        Shape of tiles for processing. [default: 2048]
    --save_thumb            To save thumb. [default: False]
    --save_mask             To save mask. [default: False]
"""
"""
Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import torch
import logging
import os
from .misc.utils import log_info
from docopt import docopt
from typing import Dict
import pdb

#-------------------------------------------------------------------------------------------------------

def main_with_args(args: Dict, sub_args: Dict, sub_cmd: str) -> None:
    gpu_list = args['gpu']
    os.environ['CUDA_VISIBLE_DEVICES'] = gpu_list

    nr_gpus = torch.cuda.device_count()
    log_info('Detect #GPUS: %d' % nr_gpus)        

    if args['model_path'] == None:
        raise Exception('A model path must be supplied as an argument with --model_path.')

    nr_types = int(args['nr_types']) if int(args['nr_types']) > 0 else None
    method_args = {
        'nr_gpus' : nr_gpus,
        'method' : {
            'model_args' : {
                'nr_types'   : nr_types,
                'mode'       : args['model_mode'],
            },
            'model_path' : args['model_path'],
        },
        'type_info_path'  : None if args['type_info_path'] == '' \
                            else args['type_info_path'],
    }

    # ***
    run_args = {
        'batch_size' : int(args['batch_size']) * max(1, nr_gpus),
        
        'nr_inference_workers' : int(args['nr_inference_workers']),
        'nr_post_proc_workers' : int(args['nr_post_proc_workers']),
    }

    if args['model_mode'] == 'fast':
        run_args['patch_input_shape'] = 256
        run_args['patch_output_shape'] = 164
    else:
        if str(args['shape']) == '270':
            run_args['patch_input_shape'] = 270
            run_args['patch_output_shape'] = 80
        elif str(args['shape']) == '518':
            run_args['patch_input_shape'] = 518
            run_args['patch_output_shape'] = 328
        else:
            assert False, 'Wrong shape at inference.'

    if sub_cmd == 'tile':
        run_args.update({
            'input_dir'      : sub_args['input_dir'],
            'output_dir'     : sub_args['output_dir'],

            'mem_usage'   : float(sub_args['mem_usage']),
            'draw_dot'    : sub_args['draw_dot'],
            'save_qupath' : sub_args['save_qupath'],
            'save_raw_map': sub_args['save_raw_map'],
        })

    if sub_cmd == 'wsi':
        run_args.update({
            'input_dir'      : sub_args['input_dir'],
            'output_dir'     : sub_args['output_dir'],
            'input_mask_dir' : sub_args['input_mask_dir'],
            'cache_path'     : sub_args['cache_path'],

            'proc_mag'       : int(sub_args['proc_mag']),
            'ambiguous_size' : int(sub_args['ambiguous_size']),
            'chunk_shape'    : int(sub_args['chunk_shape']),
            'tile_shape'     : int(sub_args['tile_shape']),
            'save_thumb'     : sub_args['save_thumb'],
            'save_mask'      : sub_args['save_mask'],
        })
    # ***
    
    if sub_cmd == 'tile':
        from .infer.tile import InferManager
        infer = InferManager(**method_args)
        infer.process_file_list(run_args)
    else:
        from .infer.wsi import InferManager
        infer = InferManager(**method_args)
        infer.process_wsi_list(run_args)

def main():
    sub_cli_dict = {'tile' : tile_cli, 'wsi' : wsi_cli}
    args = docopt(__doc__, help=False, options_first=True, 
                    version='HoVer-Net Pytorch Inference v1.0')
    sub_cmd = args.pop('<command>')
    sub_cmd_args = args.pop('<args>')

    # ! TODO: where to save logging
    logging.basicConfig(
        level=logging.INFO,
        format='|%(asctime)s.%(msecs)03d| [%(levelname)s] %(message)s',datefmt='%Y-%m-%d|%H:%M:%S',
        handlers=[
            logging.StreamHandler()
        ]
    )

    if args['--help'] and sub_cmd is not None:
        if sub_cmd in sub_cli_dict: 
            print(sub_cli_dict[sub_cmd])
        else:
            print(__doc__)
        exit()
    if args['--help'] or sub_cmd is None:
        print(__doc__)
        exit()

    sub_args = docopt(sub_cli_dict[sub_cmd], argv=sub_cmd_args, help=True)
    
    args.pop('--version')

    args = {k.replace('--', '') : v for k, v in args.items()}
    sub_args = {k.replace('--', '') : v for k, v in sub_args.items()}

    main_with_args(args, sub_args, sub_cmd)

if __name__=='__main__':
    main()


"""
Creates images with pixel value 255 at the centroids coordinates.

Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
from typing import List, Tuple
import numpy as np
import cv2
from ..utils.preprocessing import parse_path, create_dir, get_names, read_centroids
import argparse
from tqdm import tqdm


def centroids2png(centroids: List[Tuple[int, int, int]]) -> np.ndarray:
    """
    Generates a blank image with pixel value 255 at the coordinates of each centroid.

    :param centroids: A list of centroids represented as (X, Y, class) tuples.
    :type centroids: List[Tuple[int, int, int]]
    :return: A NumPy array representing the generated image.
    :rtype: np.ndarray
    """
    png = np.zeros((1024, 1024))
    for x, y, cls in centroids:
        png[x, y] = 255
    return png


def _create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--centroids-path', type=str, required=True,
                        help='Path to centroid files.')
    parser.add_argument('--output-path', type=str, required=True,
                        help='Path to save png files with points.')
    return parser


def main_with_args(args):
    centroids_path = parse_path(args.centroids_path)
    output_path = parse_path(args.output_path)
    create_dir(output_path)
    names = get_names(centroids_path, '.centroids.csv')
    for name in tqdm(names):
        centroids = read_centroids(name, centroids_path)
        png = centroids2png(centroids)
        cv2.imwrite(output_path + name + '.points.png', png)


def main():
    parser = _create_parser()
    args = parser.parse_args()
    main_with_args(args)



import json
import random

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.lines import Line2D
from termcolor import colored

from .base import BaseCallbacks
from .serialize import fig2data, serialize

# TODO: logging for all printed info on the terminal


####
class LoggingGradient(BaseCallbacks):
    """Will log per each training step."""

    def _pyplot_grad_flow(self, named_parameters):
        """Plots the gradients flowing through different layers in the net during training.
        "_pyplot_grad_flow(self.model.named_parameters())" to visualize the gradient flow.

        ! Very slow if triggered per steps because of CPU <=> GPU.

        """
        ave_grads = []
        max_grads = []
        layers = []
        for n, p in named_parameters:
            if (p.requires_grad) and ("bias" not in n):
                layers.append(n)
                ave_grads.append(p.grad.abs().mean().cpu().item())
                max_grads.append(p.grad.abs().max().cpu().item())
        fig = plt.figure(figsize=(10, 10))
        plt.bar(np.arange(len(max_grads)), max_grads, alpha=0.1, lw=1, color="c")
        plt.bar(np.arange(len(max_grads)), ave_grads, alpha=0.1, lw=1, color="b")
        plt.hlines(0, 0, len(ave_grads) + 1, lw=2, color="k")
        plt.xticks(range(0, len(ave_grads), 1), layers, rotation="vertical")
        plt.xlim(left=0, right=len(ave_grads))
        # zoom in on the lower gradient regions
        plt.xlabel("Layers")
        plt.ylabel("average gradient")
        plt.title("Gradient flow")
        plt.grid(True)
        plt.legend(
            [
                Line2D([0], [0], color="c", lw=4),
                Line2D([0], [0], color="b", lw=4),
                Line2D([0], [0], color="k", lw=4),
            ],
            ["max-gradient", "mean-gradient", "zero-gradient"],
        )
        fig = np.transpose(fig2data(fig), axes=[2, 0, 1])  # HWC => CHW
        plt.close()
        return fig

    def run(self, state, event):

        if random.random() > 0.05:
            return
        curr_step = state.curr_global_step

        # logging the grad of all trainable parameters
        tfwriter = state.log_info["tfwriter"]
        run_info = state.run_info
        for net_name, net_info in run_info.items():
            netdesc = net_info["desc"].module
            for param_name, param in netdesc.named_parameters():
                param_grad = param.grad
                # TODO: sync test None or epislon for pytorch 1.4 vs 1.5
                if param_grad is None:
                    continue
                tfwriter.add_histogram(
                    "%s_grad/%s" % (net_name, param_name),
                    param_grad.detach().cpu().numpy().flatten(),
                    global_step=curr_step,
                )  # ditribute into 10 bins (np default)
                tfwriter.add_histogram(
                    "%s_para/%s" % (net_name, param_name),
                    param.detach().cpu().numpy().flatten(),
                    global_step=curr_step,
                )  # ditribute into 10 bins (np default)
        return


####
class LoggingEpochOutput(BaseCallbacks):
    """Must declare save dir first in the shared global state of the attached engine."""

    def __init__(self, per_n_epoch=1):
        super().__init__()
        self.per_n_epoch = per_n_epoch

    def run(self, state, event):

        # only logging every n epochs also
        if state.curr_epoch % self.per_n_epoch != 0:
            return

        # TODO: rename to differentiate global vs local epoch
        if state.global_state is not None:
            current_epoch = str(state.global_state.curr_epoch)
        else:
            current_epoch = str(state.curr_epoch)

        output = state.tracked_step_output

        def get_serializable_values(output_format):
            log_dict = {}
            # get type and variable that is serializable
            # to console or other logging format (json, tensorboard)
            for variable_type, variable_dict in output.items():
                for value_name, value in variable_dict.items():
                    value_name = "%s-%s" % (state.attached_engine_name, value_name)
                    new_format = serialize(value, variable_type, output_format)
                    if new_format is not None:
                        log_dict[value_name] = new_format
            return log_dict

        # * Serialize to Console
        # align the console print output
        formatted_values = get_serializable_values("console")
        max_length = len(max(formatted_values.keys(), key=len))
        for value_name, value_text in formatted_values.items():
            value_name = colored(value_name.ljust(max_length), "green")
            print("------%s : %s" % (value_name, value_text))

        # TODO: [CRITICAL] fix passing this between engine
        # if not state.logging: return

        # * Serialize to JSON file
        stat_dict = get_serializable_values("json")
        # json stat log file, update and overwrite
        with open(state.log_info["json_file"]) as json_file:
            json_data = json.load(json_file)

        if current_epoch in json_data:
            old_stat_dict = json_data[current_epoch]
            stat_dict.update(old_stat_dict)
        current_epoch_dict = {current_epoch: stat_dict}
        json_data.update(current_epoch_dict)

        # TODO: may corrupt
        with open(state.log_info["json_file"], "w") as json_file:
            json.dump(json_data, json_file)

        # * Serialize to Tensorboard
        tfwriter = state.log_info["tfwriter"]
        formatted_values = get_serializable_values("tensorboard")
        # ! may need to flush to force update
        for value_name, value in formatted_values.items():
            # TODO: dynamically call this
            if value[0] == "scalar":
                tfwriter.add_scalar(value_name, value[1], current_epoch)
            elif value[0] == "image":
                tfwriter.add_image(
                    value_name, value[1], current_epoch, dataformats="HWC"
                )
        # tfwriter.flush()

        return



"""
Auxiliar script to compute how imbalanced the dataset is.

Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import argparse
from ..utils.read_nodes import read_all_nodes
import os
import numpy as np

FILE_DIR = os.path.dirname(os.path.abspath(__file__))

parser = argparse.ArgumentParser()
parser.add_argument('--graph-dir', type=str, required=True,
                    help='Folder containing .graph.csv files.')


if __name__ == '__main__':
    args = parser.parse_args()
    GRAPH_DIR = args.graph_dir

    X, y = read_all_nodes(GRAPH_DIR, os.listdir(GRAPH_DIR))
    print('Number of tumoral cells:', np.sum(y == 1))
    print('Number of non-tumoral cells:', np.sum(y == 0))



import math
import numpy as np
import torch
import torch.utils.data as data


####
class SerializeFileList(data.IterableDataset):
    """Read a single file as multiple patches of same shape, perform the padding beforehand."""

    def __init__(self, img_list, patch_info_list, patch_size, preproc=None):
        super().__init__()
        self.patch_size = patch_size

        self.img_list = img_list
        self.patch_info_list = patch_info_list

        self.worker_start_img_idx = 0
        # * for internal worker state
        self.curr_img_idx = 0
        self.stop_img_idx = 0
        self.curr_patch_idx = 0
        self.stop_patch_idx = 0
        self.preproc = preproc
        return

    def __iter__(self):
        worker_info = torch.utils.data.get_worker_info()
        if worker_info is None:  # single-process data loading, return the full iterator
            self.stop_img_idx = len(self.img_list)
            self.stop_patch_idx = len(self.patch_info_list)
            return self
        else:  # in a worker process so split workload, return a reduced copy of self
            per_worker = len(self.patch_info_list) / float(worker_info.num_workers)
            per_worker = int(math.ceil(per_worker))

            global_curr_patch_idx = worker_info.id * per_worker
            global_stop_patch_idx = global_curr_patch_idx + per_worker
            self.patch_info_list = self.patch_info_list[
                global_curr_patch_idx:global_stop_patch_idx
            ]
            self.curr_patch_idx = 0
            self.stop_patch_idx = len(self.patch_info_list)
            # * check img indexer, implicit protocol in infer.py
            global_curr_img_idx = self.patch_info_list[0][-1]
            global_stop_img_idx = self.patch_info_list[-1][-1] + 1
            self.worker_start_img_idx = global_curr_img_idx
            self.img_list = self.img_list[global_curr_img_idx:global_stop_img_idx]
            self.curr_img_idx = 0
            self.stop_img_idx = len(self.img_list)
            return self  # does it mutate source copy?

    def __next__(self):

        if self.curr_patch_idx >= self.stop_patch_idx:
            raise StopIteration  # when there is nothing more to yield
        patch_info = self.patch_info_list[self.curr_patch_idx]
        img_ptr = self.img_list[patch_info[-1] - self.worker_start_img_idx]
        patch_data = img_ptr[
            patch_info[0] : patch_info[0] + self.patch_size,
            patch_info[1] : patch_info[1] + self.patch_size,
        ]
        self.curr_patch_idx += 1
        if self.preproc is not None:
            patch_data = self.preproc(patch_data)
        return patch_data, patch_info


####
class SerializeArray(data.Dataset):
    def __init__(self, mmap_array_path, patch_info_list, patch_size, preproc=None):
        super().__init__()
        self.patch_size = patch_size

        # use mmap as intermediate sharing, else variable will be duplicated
        # accross torch worker => OOM error, open in read only mode
        self.image = np.load(mmap_array_path, mmap_mode="r")

        self.patch_info_list = patch_info_list
        self.preproc = preproc
        return

    def __len__(self):
        return len(self.patch_info_list)

    def __getitem__(self, idx):
        patch_info = self.patch_info_list[idx]
        patch_data = self.image[
            patch_info[0] : patch_info[0] + self.patch_size[0],
            patch_info[1] : patch_info[1] + self.patch_size[1],
        ]
        if self.preproc is not None:
            patch_data = self.preproc(patch_data)
        return patch_data, patch_info



import cv2
import matplotlib
import numpy as np
from matplotlib import pyplot as plt
import io
import pandas as pd

# * syn where to set this
# must use 'Agg' to plot out onto image
matplotlib.use("Agg")

####
def fig2data(fig, dpi=180):
    """Convert a Matplotlib figure to a 4D numpy array with RGBA channels and return it.
    
    Args:
        fig: a matplotlib figure
    
    Return: a numpy 3D array of RGBA values

    """
    buf = io.BytesIO()
    fig.savefig(buf, format="png", dpi=dpi)
    buf.seek(0)
    img_arr = np.frombuffer(buf.getvalue(), dtype=np.uint8)
    buf.close()
    img = cv2.imdecode(img_arr, 1)
    img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    return img


####
class _Scalar(object):
    @staticmethod
    def to_console(value):
        return "%0.5f" % value

    @staticmethod
    def to_json(value):
        return value

    @staticmethod
    def to_tensorboard(value):
        return "scalar", value


####
class _ConfusionMatrix(object):
    @staticmethod
    def to_console(value):
        value = pd.DataFrame(value)
        value.index.name = "True"
        value.columns.name = "Pred"
        formatted_value = value.to_string()
        return "\n" + formatted_value

    @staticmethod
    def to_json(value):
        value = pd.DataFrame(value)
        value.index.name = "True"
        value.columns.name = "Pred"
        value = value.unstack().rename("value").reset_index()
        value = pd.Series({"conf_mat": value})
        formatted_value = value.to_json(orient="records")
        return formatted_value

    @staticmethod
    def to_tensorboard(value):
        def plot_confusion_matrix(
            cm, target_names, title="Confusion matrix", cmap=None, normalize=False
        ):
            """given a sklearn confusion matrix (cm), make a nice plot.

            Args:
                cm:           confusion matrix from sklearn.metrics.confusion_matrix

                target_names: given classification classes such as [0, 1, 2]
                            the class names, for example: ['high', 'medium', 'low']

                title:        the text to display at the top of the matrix

                cmap:         the gradient of the values displayed from matplotlib.pyplot.cm
                            see http://matplotlib.org/examples/color/colormaps_reference.html
                            plt.get_cmap('jet') or plt.cm.Blues

                normalize:    If False, plot the raw numbers
                            If True, plot the proportions

            Usage
            -----
            plot_confusion_matrix(cm           = cm,                  # confusion matrix created by
                                                                    # sklearn.metrics.confusion_matrix
                                normalize    = True,                # show proportions
                                target_names = y_labels_vals,       # list of names of the classes
                                title        = best_estimator_name) # title of graph

            Citiation
            ---------
            http://scikit-learn.org/stable/auto_examples/model_selection/plot_confusion_matrix.html

            """
            import matplotlib.pyplot as plt
            import numpy as np
            import itertools

            accuracy = np.trace(cm) / np.sum(cm).astype("float")
            misclass = 1 - accuracy

            if cmap is None:
                cmap = plt.get_cmap("Blues")

            plt.figure(figsize=(8, 6))
            plt.imshow(cm, interpolation="nearest", cmap=cmap)
            plt.title(title)
            plt.colorbar()

            if target_names is not None:
                tick_marks = np.arange(len(target_names))
                plt.xticks(tick_marks, target_names, rotation=45)
                plt.yticks(tick_marks, target_names)

            if normalize:
                cm = cm.astype("float") / cm.sum(axis=1)[:, np.newaxis]

            thresh = cm.max() / 1.5 if normalize else cm.max() / 2
            for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
                if normalize:
                    plt.text(
                        j,
                        i,
                        "{:0.4f}".format(cm[i, j]),
                        horizontalalignment="center",
                        color="white" if cm[i, j] > thresh else "black",
                    )
                else:
                    plt.text(
                        j,
                        i,
                        "{:,}".format(cm[i, j]),
                        horizontalalignment="center",
                        color="white" if cm[i, j] > thresh else "black",
                    )

            plt.tight_layout()
            plt.ylabel("True label")
            plt.xlabel(
                "Predicted label\naccuracy={:0.4f}; misclass={:0.4f}".format(
                    accuracy, misclass
                )
            )

        plot_confusion_matrix(value, ["0", "1"])
        img = fig2data(plt.gcf())
        plt.close()
        return "image", img


####
class _Image(object):
    @staticmethod
    def to_console(value):
        # TODO: add warn for not possible or sthg here
        return None

    @staticmethod
    def to_json(value):
        # TODO: add warn for not possible or sthg here
        return None

    @staticmethod
    def to_tensorboard(value):
        # TODO: add method
        return "image", value


__converter_dict = {"scalar": _Scalar, "conf_mat": _ConfusionMatrix, "image": _Image}


####
def serialize(value, input_format, output_format):
    converter = __converter_dict[input_format]
    if output_format == "console":
        return converter.to_console(value)
    elif output_format == "json":
        return converter.to_json(value)
    elif output_format == "tensorboard":
        return converter.to_tensorboard(value)
    else:
        assert False, "Unknown format"



import torch.optim as optim

from ...run_utils.callbacks.base import (
    AccumulateRawOutput,
    PeriodicSaver,
    ProcessAccumulatedRawOutput,
    ScalarMovingAverage,
    ScheduleLr,
    TrackLr,
    VisualizeOutput,
    TriggerEngine,
)
from ...run_utils.callbacks.logging import LoggingEpochOutput
from ...run_utils.engine import Events

from .targets import gen_targets, prep_sample
from .net_desc import create_model
from .run_desc import proc_valid_step_output, train_step, valid_step, viz_step_output


def get_config(nr_type, mode):
    return {
        # ------------------------------------------------------------------
        # ! All phases have the same number of run engine
        # phases are run sequentially from index 0 to N
        "phase_list": [
            {
                "run_info": {
                    # may need more dynamic for each network
                    "net": {
                        "use_cpu": False,
                        "desc": lambda: create_model(
                            input_ch=3, nr_types=nr_type, 
                            freeze=True, mode=mode
                        ),
                        "optimizer": [
                            optim.Adam,
                            {  # should match keyword for parameters within the optimizer
                                "lr": 1.0e-5,  # initial learning rate,
                                "betas": (0.9, 0.999),
                            },
                        ],
                        # learning rate scheduler
                        "lr_scheduler": lambda x: optim.lr_scheduler.StepLR(x, 25),
                        "extra_info": {
                            "loss": {
                                "np": {"bce": 1, "dice": 1},
                                "hv": {"mse": 1, "msge": 1},
                                "tp": {"bce": 1, "dice": 1},
                            },
                        },
                        # path to load, -1 to auto load checkpoint from previous phase,
                        # None to start from scratch
                        # "pretrained": -1,
                        "pretrained": "Useless crap",# "/mnt/gpid08/users/jose.perez.cano/checkpoints/4/01/net_epoch=50.tar",
                        # 'pretrained': None,
                    },
                },
                "target_info": {"gen": (gen_targets, {}), "viz": (prep_sample, {})},
                "batch_size": {"train": 4, "valid": 2,},  # engine name : value
                "nr_epochs": 50,
            },
            {
                "run_info": {
                    # may need more dynamic for each network
                    "net": {
                        "use_cpu": False,
                        "desc": lambda: create_model(
                            input_ch=3, nr_types=nr_type, 
                            freeze=False, mode=mode
                        ),
                        "optimizer": [
                            optim.Adam,
                            {  # should match keyword for parameters within the optimizer
                                "lr": 1.0e-4,  # initial learning rate,
                                "betas": (0.9, 0.999),
                            },
                        ],
                        # learning rate scheduler
                        "lr_scheduler": lambda x: optim.lr_scheduler.StepLR(x, 25),
                        "extra_info": {
                            "loss": {
                                "np": {"bce": 1, "dice": 1},
                                "hv": {"mse": 1, "msge": 1},
                                "tp": {"bce": 1, "dice": 1},
                            },
                        },
                        # path to load, -1 to auto load checkpoint from previous phase,
                        # None to start from scratch
                        "pretrained": -1,
                    },
                },
                "target_info": {"gen": (gen_targets, {}), "viz": (prep_sample, {})},
                "batch_size": {"train": 2, "valid": 2,}, # batch size per gpu
                "nr_epochs": 50,
            },
        ],
        # ------------------------------------------------------------------
        "run_engine": {
            "train": {
                "dataset": "",  # whats about compound dataset ?
                "nr_procs": 1,  # number of threads for dataloader
                "run_step": train_step,  # TODO: function name or function variable ?
                "reset_per_run": False,
                # callbacks are run according to the list order of the event
                "callbacks": {
                    Events.STEP_COMPLETED: [
                        # LoggingGradient(), # TODO: very slow, may be due to back forth of tensor/numpy ?
                        ScalarMovingAverage(),
                    ],
                    Events.EPOCH_COMPLETED: [
                        TrackLr(),
                        PeriodicSaver(per_n_epoch=5),
                        VisualizeOutput(viz_step_output),
                        LoggingEpochOutput(),
                        TriggerEngine("valid"),
                        ScheduleLr(),
                    ],
                },
            },
            "valid": {
                "dataset": "",  # whats about compound dataset ?
                "nr_procs": 1,  # number of threads for dataloader
                "run_step": valid_step,
                "reset_per_run": True,  # * to stop aggregating output etc. from last run
                # callbacks are run according to the list order of the event
                "callbacks": {
                    Events.STEP_COMPLETED: [AccumulateRawOutput(),],
                    Events.EPOCH_COMPLETED: [
                        # TODO: is there way to preload these ?
                        ProcessAccumulatedRawOutput(
                            lambda a: proc_valid_step_output(a, nr_types=nr_type)
                        ),
                        LoggingEpochOutput(),
                    ],
                },
            },
        },
    }



"""
Module to create graph from nodes.

Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
from itertools import tee
from typing import Tuple, List, Optional, Callable, Any
import os
import numpy as np
from ..utils.preprocessing import get_names
from ..utils.nearest import generate_tree
from ..utils.classification import fit_column_normalizer
from ..utils.read_nodes import read_node_matrix
import torch
from torch.utils.data import Dataset
import dgl
from sklearn.preprocessing import Normalizer


class GraphDataset(Dataset):
    """
    Torch Dataset to load graphs from .nodes.csv files.

    Generated graph is in DGL format, with node attributes in .ndata
    and edge attributes in .edata.

    Graph edges are generated on the fly.
    """
    def __init__(
            self, node_dir: str, max_dist: float, max_degree: int,
            files: Optional[List[str]] = None,
            transform: Optional[Callable[[np.ndarray], np.ndarray]] = None,
            column_normalize: Optional[bool] = False,
            row_normalize: Optional[bool] = False,
            normalizers: Optional[Tuple[Any]] = None,
            return_names: Optional[bool] = False,
            is_inference: Optional[bool] = False,
            remove_prior: Optional[bool] = False,
            remove_morph: Optional[bool] = False,
            ):
        """
        node_dir: Path to .nodes.csv files.
        max_dist: Maximum distance to consider two nodes as neighbours.
        max_degree: Maximum degree for each node.
        files: List of names to include in the dataset. If None all names are included.
        column_normalize: Whether to subtract mean and divide by standard deviation each feature.
        row_normalize: Whether to apply row normalization. Reference: https://pytorch-geometric.readthedocs.io/en/latest/_modules/torch_geometric/transforms/normalize_features.html#NormalizeFeatures
        normalizers: sklearn tuple of objects with transform method.
        """
        super().__init__()
        self.node_dir = node_dir
        if files is not None:
            self.node_names = set(get_names(node_dir, '.nodes.csv'))
            self.node_names = sorted(list(self.node_names.intersection(set(files))))
        else:
            self.node_names = sorted(get_names(node_dir, '.nodes.csv'))
        self.max_dist = max_dist
        self.max_degree = max_degree
        self.transform = transform
        self.column_normalize = column_normalize
        self.row_normalize = row_normalize
        self.normalizers = normalizers
        self.remove_prior = remove_prior
        self.remove_morph = remove_morph
        self.initialize_normalizers()
        self.return_names = return_names
        self.is_inference = is_inference

    def __getitem__(self, idx):
        file_name = self.node_names[idx] + '.nodes.csv'
        X, y, xx, yy = read_node_matrix(
            os.path.join(self.node_dir, file_name), return_coordinates=True, return_class=not self.is_inference,
            remove_prior=self.remove_prior, remove_morph=self.remove_morph
            )
        if self.column_normalize:
            X = self.col_sc.transform(X)
        if self.row_normalize:
            X = self.row_sc.transform(X)
        if self.normalizers is not None:
            for normalizer in self.normalizers:
                X = normalizer.transform(X)
        if self.transform is not None:
            X = self.transform(X)
        source, dest, dists = GraphDataset.create_edges(xx, yy, self.max_degree, self.max_dist)
        g = dgl.graph((source, dest), num_nodes=len(X))
        g.ndata['X'] = torch.tensor(X, dtype=torch.float32)
        if not self.is_inference:
            g.ndata['y'] = torch.tensor(y, dtype=torch.long)
        g.edata['dist'] = torch.tensor(dists, dtype=torch.float32).reshape((-1, 1))
        if self.return_names:
            return g, file_name
        return g

    def __len__(self):
        return len(self.node_names)

    @staticmethod
    def create_edges(
            xx: List[float], yy: List[float],
            max_degree: int, threshold: float
            ) -> Tuple[List[int], List[int]]:
        """
        Creates edges between nearby nodes.

        xx: X coordinates of nodes.
        yy: Y coordinates of nodes.
        max_degree: Maximum degree for each node.
        threshold: Maximum distance to look at.

        Returns
        source: List of source nodes id.
        dest: List of destination nodes id.
        distances: Distances between nodes in edges.
        """
        tree = generate_tree(zip(xx, yy))
        source, dest, distances = [], [], []
        for i, (x, y) in enumerate(zip(xx, yy)):
            dists, idx = tree.query((x, y), k=max_degree, distance_upper_bound=threshold)
            tmp = list(filter(lambda x: x[0] > 1e-10 and x[1] < len(xx), zip(dists, idx)))
            tmp1, tmp2 = tee(tmp)
            dists, idx = list(x[0] for x in tmp1), list(x[1] for x in tmp2)
            source.extend([i for _ in range(len(idx))])
            dest.extend(idx)
            distances.extend(dists)
        return source, dest, distances

    def initialize_normalizers(self):
        """
        Fits normalizers for later use and also checks they contain transform method.
        """
        if self.normalizers is not None:
            for normalizer in self.normalizers:
                assert callable(getattr(normalizer, "transform", None)), \
                    'Normalizers provided must have transform method.'
        if self.column_normalize:
            self.col_sc = fit_column_normalizer(self.node_dir, self.node_names, remove_morph=self.remove_morph, remove_prior=self.remove_prior)
            assert callable(getattr(self.col_sc, "transform", None)), \
                'Error loading column normalizer.'
        if self.row_normalize:
            self.row_sc = Normalizer(norm='l1')
            assert callable(getattr(self.row_sc, "transform", None)), \
                'Error loading row normalizer.'

    def get_normalizers(self) -> Tuple[Any]:
        """
        Returns a tuple with all the normalizers in the order they are used.
        """
        if self.column_normalize and self.row_normalize and self.normalizers is not None:
            return [self.col_sc, self.row_sc, *self.normalizers]
        if self.column_normalize and self.row_normalize:
            return [self.col_sc, self.row_sc]
        if self.column_normalize and self.normalizers is not None:
            return [self.col_sc, *self.normalizers]
        if self.row_normalize and self.normalizers is not None:
            return [self.row_sc, *self.normalizers]
        if self.column_normalize:
            return [self.col_sc]
        if self.row_normalize and self.normalizers is not None:
            return [self.row_sc, *self.normalizers]
        if self.normalizers is not None:
            return self.normalizers



import gradio as gr
import os
import numpy as np
import shutil
import cv2
import json
from tumourkit.segmentation import hov_infer
from tumourkit.preprocessing import geojson2pngcsv_main, png2graph_main, hovernet2geojson_main, graph2centroids_main, centroidspng2csv_main, pngcsv2geojson_main
from tumourkit.postprocessing import join_hovprob_graph_main, draw_cells_main
from tumourkit.utils.preprocessing import create_dir
from tumourkit.classification import infer_gnn
import requests
from tqdm import tqdm
from bs4 import BeautifulSoup
from argparse import Namespace
import logging
from logging import Logger
import argparse
from typing import Tuple, Optional


APP_DIR = os.path.dirname(os.path.abspath(__file__))


def download_file(url: str, filename: str):
    """
    Downloads a file from a specified URL and saves it to the given filename.

    :param url: The URL of the file to be downloaded.
    :type url: str

    :param filename: The name of the file to be saved.
    :type filename: str
    """
    response = requests.get(url, stream=True)
    file_size = int(response.headers.get('Content-Length', 0))
    showname = '...' + filename[-20:] if len(filename) > 20 else filename
    progress = tqdm(response.iter_content(1024), f"Downloading {showname}", total=file_size, unit="B", unit_scale=True, unit_divisor=1024)
    with open(filename, 'wb') as f:
        for data in progress.iterable:
            f.write(data)
            progress.update(len(data))
    progress.close()


def download_folder(url_folder: str, url_files: str, dirname: str):
    """
    Downloads multiple files from a specified folder URL and saves them to the given directory.

    :param url_folder: The URL of the folder containing the files to be downloaded.
    :type url_folder: str

    :param url_files: The files within the folder.
    :type url_files: str

    :param dirname: The name of the directory where the files will be saved.
    :type dirname: str
    """
    response = requests.get(url_folder)
    html_content = response.content
    soup = BeautifulSoup(html_content, 'html.parser')
    file_list = [span.get_text() for span in soup.find_all('span') if span.get_text().startswith('best_')]

    for file in file_list:
        filename = os.path.join(dirname, file)
        url = url_files + file
        response = requests.get(url, stream=True)
        file_size = int(response.headers.get('Content-Length', 0))
        showname = '...' + filename[-20:] if len(filename) > 20 else filename
        progress = tqdm(response.iter_content(1024), f"Downloading {showname}", total=file_size, unit="B", unit_scale=True, unit_divisor=1024)
        with open(filename, 'wb') as f:
            for data in progress.iterable:
                f.write(data)
                progress.update(len(data))
        progress.close()


def download_models_if_needed(hov_dataset: str, hov_model: str, gnn_model: str, weights_dir: str):
    """
    Downloads necessary models and files if they do not exist in the specified directory.

    :param hov_dataset: The name of the Hovernet dataset.
    :type hov_dataset: str

    :param hov_model: The name of the Hovernet model.
    :type hov_model: str

    :param gnn_model: The name of the GNN (Graph Neural Network) model.
    :type gnn_model: str

    :param weights_dir: The directory path where the models and files will be saved.
    :type weights_dir: str
    """
    if not os.path.exists(os.path.join(weights_dir, hov_dataset, hov_model + '.tar')):
        os.makedirs(os.path.join(weights_dir, hov_dataset), exist_ok=True)
        url = f'https://huggingface.co/Jerry-Master/Hovernet-plus-Graphs/resolve/main/{hov_dataset}/hovernet/{hov_model}.tar'
        filename = os.path.join(weights_dir, hov_dataset, hov_model + '.tar')
        download_file(url, filename)
    if not os.path.exists(os.path.join(weights_dir, hov_dataset, 'type_info.json')):
        os.makedirs(os.path.join(weights_dir, hov_dataset), exist_ok=True)
        url = f'https://huggingface.co/Jerry-Master/Hovernet-plus-Graphs/resolve/main/{hov_dataset}/hovernet/type_info.json'
        filename = os.path.join(weights_dir, hov_dataset, 'type_info.json')
        download_file(url, filename)
    if not os.path.exists(os.path.join(weights_dir, hov_dataset, gnn_model)) \
       or len(os.listdir(os.path.join(weights_dir, hov_dataset, gnn_model))) < 3:
        os.makedirs(os.path.join(weights_dir, hov_dataset, gnn_model), exist_ok=True)
        url_folder = f'https://huggingface.co/Jerry-Master/Hovernet-plus-Graphs/tree/main/{hov_dataset}/gnn/{gnn_model}/'
        url_files = f'https://huggingface.co/Jerry-Master/Hovernet-plus-Graphs/resolve/main/{hov_dataset}/gnn/{gnn_model}/'
        dirname = os.path.join(weights_dir, hov_dataset, gnn_model)
        download_folder(url_folder, url_files, dirname)


def create_input_dir(input_image: np.ndarray, delete_prev: bool):
    """
    Creates an input directory and saves the input image in it.

    :param input_image: The input image to be saved.
    :type input_image: np.ndarray

    :param delete_prev: Flag indicating whether to delete the previous input directory if it exists.
    :type delete_prev: bool
    """
    if os.path.exists(os.path.join(APP_DIR, 'tmp')):
        if delete_prev:
            shutil.rmtree(os.path.join(APP_DIR, 'tmp'))
    input_dir = os.path.join(APP_DIR, 'tmp', 'input')
    os.makedirs(input_dir, exist_ok=True)
    cv2.imwrite(os.path.join(input_dir, 'input_image.png'), input_image[:, :, ::-1])


def run_hovernet(hov_dataset: str, hov_model: str, num_classes: int, weights_dir: str):
    """
    Runs HoverNet inference using the specified HoverNet dataset, model, and weights.

    :param hov_dataset: The name of the HoverNet dataset.
    :type hov_dataset: str

    :param hov_model: The name of the HoverNet model.
    :type hov_model: str

    :param num_classes: The number of classes in the dataset.
    :type num_classes: int

    :param weights_dir: The directory path containing the HoverNet weights.
    :type weights_dir: str
    """
    newargs = {
        'nr_types': str(num_classes + 1),
        'type_info_path': os.path.join(weights_dir, hov_dataset, 'type_info.json'),
        'gpu': '0',
        'nr_inference_workers': '0',
        'model_path': os.path.join(weights_dir, hov_dataset, hov_model + '.tar'),
        'batch_size': '10',
        'shape': hov_model[:-2] if 'FT' in hov_model else hov_model,
        'nr_post_proc_workers': '0',
        'model_mode': 'original',
        'help': False
    }
    newsubargs = {
        'input_dir': os.path.join(APP_DIR, 'tmp', 'input'),
        'output_dir': os.path.join(APP_DIR, 'tmp', 'tmp_hov'),
        'draw_dot': False,
        'save_qupath': False,
        'save_raw_map': False,
        'mem_usage': '0.2'
    }
    hov_infer(newargs, newsubargs, 'tile')


def run_posthov(num_classes: int, logger: Logger):
    """
    Runs post-processing steps on the HoverNet predictions.

    :param num_classes: The number of classes in the HoverNet predictions.
    :type num_classes: int

    :param logger: The logger object used for logging messages.
    :type logger: Logger
    """
    newargs = Namespace(
        json_dir=os.path.join(APP_DIR, 'tmp', 'tmp_hov', 'json'),
        gson_dir=os.path.join(APP_DIR, 'tmp', 'gson_hov'),
        num_classes=num_classes
    )
    hovernet2geojson_main(newargs)
    newargs = Namespace(
        gson_dir=os.path.join(APP_DIR, 'tmp', 'gson_hov'),
        png_dir=os.path.join(APP_DIR, 'tmp', 'png_hov'),
        csv_dir=os.path.join(APP_DIR, 'tmp', 'csv_hov'),
        num_classes=num_classes
    )
    geojson2pngcsv_main(newargs)
    create_dir(os.path.join(APP_DIR, 'tmp', 'graphs'))
    newargs = Namespace(
        png_dir=os.path.join(APP_DIR, 'tmp', 'png_hov'),
        orig_dir=os.path.join(APP_DIR, 'tmp', 'input'),
        output_path=os.path.join(APP_DIR, 'tmp', 'graphs', 'raw'),
        num_workers=0
    )
    png2graph_main(newargs)
    newargs = Namespace(
        json_dir=os.path.join(APP_DIR, 'tmp', 'tmp_hov', 'json'),
        graph_dir=os.path.join(APP_DIR, 'tmp', 'graphs', 'raw'),
        output_dir=os.path.join(APP_DIR, 'tmp', 'graphs', 'hovpreds'),
        num_classes=num_classes
    )
    join_hovprob_graph_main(newargs, logger)


def run_graphs(gnn_dataset: str, gnn_model: str, num_classes: int, weights_dir: str):
    """
    Runs the GNN (Graph Neural Network) on the HoverNet graph predictions.

    :param gnn_dataset: The name of the GNN dataset.
    :type gnn_dataset: str

    :param gnn_model: The name of the GNN model.
    :type gnn_model: str

    :param num_classes: The number of classes in the GNN predictions.
    :type num_classes: int

    :param weights_dir: The directory path containing the GNN weights.
    :type weights_dir: str
    """
    disable_prior = 'no-prior' in gnn_model or 'void' in gnn_model
    disable_morph_feats = 'no-morph' in gnn_model or 'void' in gnn_model
    model_name = os.listdir(os.path.join(weights_dir, gnn_dataset, gnn_model))[0]
    model_name, ext = os.path.splitext(model_name)
    newargs = Namespace(
        node_dir=os.path.join(APP_DIR, 'tmp', 'graphs', 'hovpreds'),
        output_dir=os.path.join(APP_DIR, 'tmp', 'gnn_preds'),
        weights=os.path.join(weights_dir, gnn_dataset, gnn_model, model_name + '.pth'),
        conf=os.path.join(weights_dir, gnn_dataset, gnn_model, model_name + '.json'),
        normalizers=os.path.join(weights_dir, gnn_dataset, gnn_model, model_name + '.pkl'),
        num_classes=num_classes,
        disable_prior=disable_prior,
        disable_morph_feats=disable_morph_feats,
    )
    infer_gnn(newargs)


def run_postgraphs(num_classes: int):
    """
    Runs post-processing steps on the GNN predictions.

    :param num_classes: The number of classes in the GNN predictions.
    :type num_classes: int
    """
    newargs = Namespace(
        graph_dir=os.path.join(APP_DIR, 'tmp', 'gnn_preds'),
        centroids_dir=os.path.join(APP_DIR, 'tmp', 'centroids'),
        num_classes=num_classes,
    )
    graph2centroids_main(newargs)
    newargs = Namespace(
        centroids_dir=os.path.join(APP_DIR, 'tmp', 'centroids'),
        png_dir=os.path.join(APP_DIR, 'tmp', 'png_hov'),
        csv_dir=os.path.join(APP_DIR, 'tmp', 'csv_gnn'),
    )
    centroidspng2csv_main(newargs)
    newargs = Namespace(
        png_dir=os.path.join(APP_DIR, 'tmp', 'png_hov'),
        csv_dir=os.path.join(APP_DIR, 'tmp', 'csv_gnn'),
        gson_dir=os.path.join(APP_DIR, 'tmp', 'gson_gnn'),
        num_classes=num_classes
    )
    pngcsv2geojson_main(newargs)


def create_logger() -> Logger:
    """
    Creates a logger object for logging messages.

    :return: The logger object.
    :rtype: Logger
    """
    logger = logging.getLogger('gradio')
    logger.setLevel(logging.DEBUG)
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    return logger


def overlay_outputs(hov_dataset: str, use_gnn: bool) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """
    Returns the overlayed images generated from HoverNet and optional GNN predictions.

    :param hov_dataset: The name of the HoverNet dataset.
    :type hov_dataset: str

    :param use_gnn: Flag indicating whether to include GNN predictions in the overlay.
    :type use_gnn: bool

    :return: A tuple containing the HoverNet overlay image and the GNN overlay image (if applicable).
    :rtype: Tuple[np.ndarray, Optional[np.ndarray]]
    """
    newargs = Namespace(
        orig_dir=os.path.join(APP_DIR, 'tmp', 'input'),
        png_dir=os.path.join(APP_DIR, 'tmp', 'png_hov'),
        csv_dir=os.path.join(APP_DIR, 'tmp', 'csv_hov'),
        output_dir=os.path.join(APP_DIR, 'tmp', 'overlay_hov'),
        type_info=os.path.join(APP_DIR, 'weights', hov_dataset, 'type_info.json'),
    )
    draw_cells_main(newargs)
    hov_dir = os.path.join(APP_DIR, 'tmp', 'overlay_hov')
    hov_file = os.listdir(hov_dir)[0]
    hov = cv2.imread(os.path.join(hov_dir, hov_file), -1)[:, :, ::-1]
    if use_gnn:
        newargs = Namespace(
            orig_dir=os.path.join(APP_DIR, 'tmp', 'input'),
            png_dir=os.path.join(APP_DIR, 'tmp', 'png_hov'),
            csv_dir=os.path.join(APP_DIR, 'tmp', 'csv_gnn'),
            output_dir=os.path.join(APP_DIR, 'tmp', 'overlay_gnn'),
            type_info=os.path.join(APP_DIR, 'weights', hov_dataset, 'type_info.json'),
        )
        draw_cells_main(newargs)
        gnn_dir = os.path.join(APP_DIR, 'tmp', 'overlay_gnn')
        gnn_file = os.listdir(gnn_dir)[0]
        gnn = cv2.imread(os.path.join(gnn_dir, gnn_file), -1)[:, :, ::-1]
    else:
        gnn = None
    return hov, gnn


LAST_HOV_MODEL = None
LAST_HOV_DATASET = None
LAST_IMG_HASH = None


def process_image(
        weights_dir: str,
        input_image: np.ndarray,
        hov_dataset: str,
        hov_model: str,
        gnn_model: str
        ) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """
    Processes an input image using HoverNet and optionally GNN models.

    :param weights_dir: The directory path containing the model weights. If None, default weights directory will be used.
    :type weights_dir: str

    :param input_image: The input image to be processed.
    :type input_image: np.ndarray

    :param hov_dataset: The name of the HoverNet dataset.
    :type hov_dataset: str

    :param hov_model: The name of the HoverNet model.
    :type hov_model: str

    :param gnn_model: The name of the GNN (Graph Neural Network) model.
    :type gnn_model: str

    :return: A tuple containing the HoverNet output image and the GNN output image (if applicable).
    :rtype: Tuple[np.ndarray, Optional[np.ndarray]]
    """
    # Cache information when possible
    global LAST_HOV_MODEL
    global LAST_IMG_HASH
    global LAST_HOV_DATASET
    input_hash = hash(str(input_image))
    if LAST_IMG_HASH is None:
        LAST_IMG_HASH = input_hash
    if LAST_HOV_MODEL is None or LAST_IMG_HASH is None or LAST_HOV_DATASET is None \
       or LAST_HOV_MODEL != hov_model or LAST_IMG_HASH != input_hash or LAST_HOV_DATASET != hov_dataset:
        LAST_HOV_MODEL = hov_model
        LAST_IMG_HASH = input_hash
        LAST_HOV_DATASET = hov_dataset
        delete_prev = True
    else:
        delete_prev = False

    # Monusac doesn't have graph attention
    if hov_dataset == 'monusac' and gnn_model == 'gat-full':
        gnn_model = 'None'

    if weights_dir is None:
        weights_dir = os.path.join(APP_DIR, 'weights')
    logger = create_logger()
    download_models_if_needed(hov_dataset, hov_model, gnn_model, weights_dir)
    with open(os.path.join(APP_DIR, 'weights', hov_dataset, 'type_info.json'), 'r') as f:
        type_info = json.load(f)
        num_classes = len(type_info.keys()) - 1
    create_input_dir(input_image, delete_prev)
    if delete_prev:
        run_hovernet(hov_dataset, hov_model, num_classes, weights_dir)
        run_posthov(num_classes, logger)
    use_gnn = gnn_model != 'None'
    if use_gnn:
        run_graphs(hov_dataset, gnn_model, num_classes, weights_dir)
        run_postgraphs(num_classes)
    hov, gnn = overlay_outputs(hov_dataset, use_gnn)
    return hov, gnn


def create_ui(weights_dir: str) -> gr.Interface:
    """
    Creates and returns a Gradio interface for the CNN+GNN demo.

    :param weights_dir: The directory path containing the model weights.
    :type weights_dir: str

    :return: A Gradio interface object.
    :rtype: gr.Interface
    """
    image_input = gr.Image(shape=(1024, 1024))
    hov_dataset = gr.Dropdown(choices=[
        'consep', 'monusac', 'breast', 'lung'
    ], label='Select dataset in which models were trained')
    hov_model = gr.Dropdown(choices=[
        '270', '270FT', '518', '518FT'
    ], label="Select Hovernet model")
    gnn_model = gr.Dropdown(choices=[
        'gcn-full', 'gat-full',
        'gcn-no-morph', 'gcn-no-prior', 'gcn-void',
        'None'
        ], label='Select GNN model')
    out1 = gr.Image(label='Hovernet')
    out2 = gr.Image(label='GNN')

    def func(a: np.ndarray, b: str, c: str, d: str):
        process_image(weights_dir, a, b, c, d)
    ui = gr.Interface(
        fn=func,
        inputs=[image_input, hov_dataset, hov_model, gnn_model],
        outputs=[out1, out2],
        title="CNN+GNN Demo",
        description="Upload an image to see the output of the algorithm.",
        examples=[[os.path.join(APP_DIR, 'examples', x), 'breast', '518FT', 'gcn-full'] for x in os.listdir(os.path.join(APP_DIR, 'examples'))]
    )
    return ui


def _create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--ip', type=str, default='localhost', help='Default: localhost.')
    parser.add_argument('--port', type=int, default=15000, help='Default: 15000.')
    parser.add_argument('--share', action='store_true', help='Whether to create public link for the gradio demo.')
    parser.add_argument('--weights-dir', type=str, help='Folder to save and load weights from. Leave it blank to use tumourkit internal folders.')
    return parser


def main():
    parser = _create_parser()
    args = parser.parse_args()
    ui = create_ui(args.weights_dir)
    ui.launch(server_name=args.ip, server_port=args.port, share=args.share)


if __name__ == '__main__':
    main()



from collections import OrderedDict
import cv2
import numpy as np
import openslide


class FileHandler(object):
    def __init__(self):
        """The handler is responsible for storing the processed data, parsing
        the metadata from original file, and reading it from storage. 
        """
        self.metadata = {
            ("available_mag", None),
            ("base_mag", None),
            ("vendor", None),
            ("mpp  ", None),
            ("base_shape", None),
        }
        pass

    def __load_metadata(self):
        raise NotImplementedError

    def get_full_img(self, read_mag=None, read_mpp=None):
        """Only use `read_mag` or `read_mpp`, not both, prioritize `read_mpp`.

        `read_mpp` is in X, Y format
        """
        raise NotImplementedError

    def read_region(self, coords, size):
        """Must call `prepare_reading` before hand.

        Args:
            coords (tuple): (dims_x, dims_y), 
                          top left coordinates of image region at selected 
                          `read_mag` or `read_mpp` from `prepare_reading` 
            size (tuple): (dims_x, dims_y)
                          width and height of image region at selected 
                          `read_mag` or `read_mpp` from `prepare_reading`       

        """
        raise NotImplementedError

    def get_dimensions(self, read_mag=None, read_mpp=None):
        """Will be in X, Y."""
        if read_mpp is not None:
            read_scale = (self.metadata["base_mpp"] / read_mpp)[0]
            read_mag = read_scale * self.metadata["base_mag"]
        scale = read_mag / self.metadata["base_mag"]
        # may off some pixels wrt existing mag
        return (self.metadata["base_shape"] * scale).astype(np.int32)

    def prepare_reading(self, read_mag=None, read_mpp=None, cache_path=None):
        """Only use `read_mag` or `read_mpp`, not both, prioritize `read_mpp`.

        `read_mpp` is in X, Y format.
        """
        read_lv, scale_factor = self._get_read_info(
            read_mag=read_mag, read_mpp=read_mpp
        )

        if scale_factor is None:
            self.image_ptr = None
            self.read_lv = read_lv
        else:
            np.save(cache_path, self.get_full_img(read_mag=read_mag))
            self.image_ptr = np.load(cache_path, mmap_mode="r")
        return

    def _get_read_info(self, read_mag=None, read_mpp=None):
        if read_mpp is not None:
            assert read_mpp[0] == read_mpp[1], "Not supported uneven `read_mpp`"
            read_scale = (self.metadata["base_mpp"] / read_mpp)[0]
            read_mag = read_scale * self.metadata["base_mag"]

        hires_mag = read_mag
        scale_factor = None
        if read_mag not in self.metadata["available_mag"]:
            if read_mag > self.metadata["base_mag"]:
                scale_factor = read_mag / self.metadata["base_mag"]
                hires_mag = self.metadata["base_mag"]
            else:
                mag_list = np.array(self.metadata["available_mag"])
                mag_list = np.sort(mag_list)[::-1]
                hires_mag = mag_list - read_mag
                # only use higher mag as base for loading
                hires_mag = hires_mag[hires_mag > 0]
                # use the immediate higher to save compuration
                hires_mag = mag_list[np.argmin(hires_mag)]
                scale_factor = read_mag / hires_mag

        hires_lv = self.metadata["available_mag"].index(hires_mag)
        return hires_lv, scale_factor


class OpenSlideHandler(FileHandler):
    """Class for handling OpenSlide supported whole-slide images."""

    def __init__(self, file_path):
        """file_path (string): path to single whole-slide image."""
        super().__init__()
        self.file_ptr = openslide.OpenSlide(file_path)  # load OpenSlide object
        self.metadata = self.__load_metadata()

        # only used for cases where the read magnification is different from
        self.image_ptr = None  # the existing modes of the read file
        self.read_level = None

    def __load_metadata(self):
        metadata = {}

        wsi_properties = self.file_ptr.properties
        level_0_magnification = wsi_properties[openslide.PROPERTY_NAME_OBJECTIVE_POWER]
        level_0_magnification = float(level_0_magnification)

        downsample_level = self.file_ptr.level_downsamples
        magnification_level = [level_0_magnification / lv for lv in downsample_level]

        mpp = [
            wsi_properties[openslide.PROPERTY_NAME_MPP_X],
            wsi_properties[openslide.PROPERTY_NAME_MPP_Y],
        ]
        mpp = np.array(mpp)

        metadata = [
            ("available_mag", magnification_level),  # highest to lowest mag
            ("base_mag", magnification_level[0]),
            ("vendor", wsi_properties[openslide.PROPERTY_NAME_VENDOR]),
            ("mpp  ", mpp),
            ("base_shape", np.array(self.file_ptr.dimensions)),
        ]
        return OrderedDict(metadata)

    def read_region(self, coords, size):
        """Must call `prepare_reading` before hand.

        Args:
            coords (tuple): (dims_x, dims_y), 
                          top left coordinates of image region at selected 
                          `read_mag` or `read_mpp` from `prepare_reading` 
            size (tuple): (dims_x, dims_y)
                          width and height of image region at selected 
                          `read_mag` or `read_mpp` from `prepare_reading`       

        """
        if self.image_ptr is None:
            # convert coord from read lv to lv zero
            lv_0_shape = np.array(self.file_ptr.level_dimensions[0])
            lv_r_shape = np.array(self.file_ptr.level_dimensions[self.read_lv])
            up_sample = (lv_0_shape / lv_r_shape)[0]
            new_coord = [0, 0]
            new_coord[0] = int(coords[0] * up_sample)
            new_coord[1] = int(coords[1] * up_sample)
            region = self.file_ptr.read_region(new_coord, self.read_lv, size)
        else:
            region = self.image_ptr[
                coords[1] : coords[1] + size[1], coords[0] : coords[0] + size[0]
            ]
        return np.array(region)[..., :3]

    def get_full_img(self, read_mag=None, read_mpp=None):
        """Only use `read_mag` or `read_mpp`, not both, prioritize `read_mpp`.

        `read_mpp` is in X, Y format.
        """

        read_lv, scale_factor = self._get_read_info(
            read_mag=read_mag, read_mpp=read_mpp
        )

        read_size = self.file_ptr.level_dimensions[read_lv]

        wsi_img = self.file_ptr.read_region((0, 0), read_lv, read_size)
        wsi_img = np.array(wsi_img)[..., :3]  # remove alpha channel
        if scale_factor is not None:
            # now rescale then return
            if scale_factor > 1.0:
                interp = cv2.INTER_CUBIC
            else:
                interp = cv2.INTER_LINEAR
            wsi_img = cv2.resize(
                wsi_img, (0, 0), fx=scale_factor, fy=scale_factor, interpolation=interp
            )
        return wsi_img


def get_file_handler(path, backend):
    if backend in [
            '.svs', '.tif', 
            '.vms', '.vmu', '.ndpi',
            '.scn', '.mrxs', '.tiff',
            '.svslide',
            '.bif',
            ]:
        return OpenSlideHandler(path)
    else:
        assert False, "Unknown WSI format `%s`" % backend




import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from collections import OrderedDict
from .utils import crop_to_shape


####
class Net(nn.Module):
    """ A base class provides a common weight initialisation scheme."""

    def weights_init(self):
        for m in self.modules():
            classname = m.__class__.__name__

            # ! Fixed the type checking
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")

            if "norm" in classname.lower():
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

            if "linear" in classname.lower():
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

    def forward(self, x):
        return x


####
class TFSamepaddingLayer(nn.Module):
    """To align with tf `same` padding. 
    
    Putting this before any conv layer that need padding
    Assuming kernel has Height == Width for simplicity
    """

    def __init__(self, ksize, stride):
        super(TFSamepaddingLayer, self).__init__()
        self.ksize = ksize
        self.stride = stride

    def forward(self, x):
        if x.shape[2] % self.stride == 0:
            pad = max(self.ksize - self.stride, 0)
        else:
            pad = max(self.ksize - (x.shape[2] % self.stride), 0)

        if pad % 2 == 0:
            pad_val = pad // 2
            padding = (pad_val, pad_val, pad_val, pad_val)
        else:
            pad_val_start = pad // 2
            pad_val_end = pad - pad_val_start
            padding = (pad_val_start, pad_val_end, pad_val_start, pad_val_end)
        # print(x.shape, padding)
        x = F.pad(x, padding, "constant", 0)
        # print(x.shape)
        return x


####
class DenseBlock(Net):
    """Dense Block as defined in:

    Huang, Gao, Zhuang Liu, Laurens Van Der Maaten, and Kilian Q. Weinberger. 
    "Densely connected convolutional networks." In Proceedings of the IEEE conference 
    on computer vision and pattern recognition, pp. 4700-4708. 2017.

    Only performs `valid` convolution.

    """

    def __init__(self, in_ch, unit_ksize, unit_ch, unit_count, split=1):
        super(DenseBlock, self).__init__()
        assert len(unit_ksize) == len(unit_ch), "Unbalance Unit Info"

        self.nr_unit = unit_count
        self.in_ch = in_ch
        self.unit_ch = unit_ch

        # ! For inference only so init values for batchnorm may not match tensorflow
        unit_in_ch = in_ch
        self.units = nn.ModuleList()
        for idx in range(unit_count):
            self.units.append(
                nn.Sequential(
                    OrderedDict(
                        [
                            ("preact_bna/bn", nn.BatchNorm2d(unit_in_ch, eps=1e-5)),
                            ("preact_bna/relu", nn.ReLU(inplace=True)),
                            (
                                "conv1",
                                nn.Conv2d(
                                    unit_in_ch,
                                    unit_ch[0],
                                    unit_ksize[0],
                                    stride=1,
                                    padding=0,
                                    bias=False,
                                ),
                            ),
                            ("conv1/bn", nn.BatchNorm2d(unit_ch[0], eps=1e-5)),
                            ("conv1/relu", nn.ReLU(inplace=True)),
                            # ('conv2/pool', TFSamepaddingLayer(ksize=unit_ksize[1], stride=1)),
                            (
                                "conv2",
                                nn.Conv2d(
                                    unit_ch[0],
                                    unit_ch[1],
                                    unit_ksize[1],
                                    groups=split,
                                    stride=1,
                                    padding=0,
                                    bias=False,
                                ),
                            ),
                        ]
                    )
                )
            )
            unit_in_ch += unit_ch[1]

        self.blk_bna = nn.Sequential(
            OrderedDict(
                [
                    ("bn", nn.BatchNorm2d(unit_in_ch, eps=1e-5)),
                    ("relu", nn.ReLU(inplace=True)),
                ]
            )
        )

    def out_ch(self):
        return self.in_ch + self.nr_unit * self.unit_ch[-1]

    def forward(self, prev_feat):
        for idx in range(self.nr_unit):
            new_feat = self.units[idx](prev_feat)
            prev_feat = crop_to_shape(prev_feat, new_feat)
            prev_feat = torch.cat([prev_feat, new_feat], dim=1)
        prev_feat = self.blk_bna(prev_feat)

        return prev_feat


####
class ResidualBlock(Net):
    """Residual block as defined in:

    He, Kaiming, Xiangyu Zhang, Shaoqing Ren, and Jian Sun. "Deep residual learning 
    for image recognition." In Proceedings of the IEEE conference on computer vision 
    and pattern recognition, pp. 770-778. 2016.

    """

    def __init__(self, in_ch, unit_ksize, unit_ch, unit_count, stride=1):
        super(ResidualBlock, self).__init__()
        assert len(unit_ksize) == len(unit_ch), "Unbalance Unit Info"

        self.nr_unit = unit_count
        self.in_ch = in_ch
        self.unit_ch = unit_ch

        # ! For inference only so init values for batchnorm may not match tensorflow
        unit_in_ch = in_ch
        self.units = nn.ModuleList()
        for idx in range(unit_count):
            unit_layer = [
                ("preact/bn", nn.BatchNorm2d(unit_in_ch, eps=1e-5)),
                ("preact/relu", nn.ReLU(inplace=True)),
                (
                    "conv1",
                    nn.Conv2d(
                        unit_in_ch,
                        unit_ch[0],
                        unit_ksize[0],
                        stride=1,
                        padding=0,
                        bias=False,
                    ),
                ),
                ("conv1/bn", nn.BatchNorm2d(unit_ch[0], eps=1e-5)),
                ("conv1/relu", nn.ReLU(inplace=True)),
                (
                    "conv2/pad",
                    TFSamepaddingLayer(
                        ksize=unit_ksize[1], stride=stride if idx == 0 else 1
                    ),
                ),
                (
                    "conv2",
                    nn.Conv2d(
                        unit_ch[0],
                        unit_ch[1],
                        unit_ksize[1],
                        stride=stride if idx == 0 else 1,
                        padding=0,
                        bias=False,
                    ),
                ),
                ("conv2/bn", nn.BatchNorm2d(unit_ch[1], eps=1e-5)),
                ("conv2/relu", nn.ReLU(inplace=True)),
                (
                    "conv3",
                    nn.Conv2d(
                        unit_ch[1],
                        unit_ch[2],
                        unit_ksize[2],
                        stride=1,
                        padding=0,
                        bias=False,
                    ),
                ),
            ]
            # * has bna to conclude each previous block so
            # * must not put preact for the first unit of this block
            unit_layer = unit_layer if idx != 0 else unit_layer[2:]
            self.units.append(nn.Sequential(OrderedDict(unit_layer)))
            unit_in_ch = unit_ch[-1]

        if in_ch != unit_ch[-1] or stride != 1:
            self.shortcut = nn.Conv2d(in_ch, unit_ch[-1], 1, stride=stride, bias=False)
        else:
            self.shortcut = None

        self.blk_bna = nn.Sequential(
            OrderedDict(
                [
                    ("bn", nn.BatchNorm2d(unit_in_ch, eps=1e-5)),
                    ("relu", nn.ReLU(inplace=True)),
                ]
            )
        )

        # print(self.units[0])
        # print(self.units[1])
        # exit()

    def out_ch(self):
        return self.unit_ch[-1]

    def forward(self, prev_feat, freeze=False):
        if self.shortcut is None:
            shortcut = prev_feat
        else:
            shortcut = self.shortcut(prev_feat)

        for idx in range(0, len(self.units)):
            new_feat = prev_feat
            if self.training:
                with torch.set_grad_enabled(not freeze):
                    new_feat = self.units[idx](new_feat)
            else:
                new_feat = self.units[idx](new_feat)
            prev_feat = new_feat + shortcut
            shortcut = prev_feat
        feat = self.blk_bna(prev_feat)
        return feat


####
class UpSample2x(nn.Module):
    """Upsample input by a factor of 2.
    
    Assume input is of NCHW, port FixedUnpooling from TensorPack.
    """

    def __init__(self):
        super(UpSample2x, self).__init__()
        # correct way to create constant within module
        self.register_buffer(
            "unpool_mat", torch.from_numpy(np.ones((2, 2), dtype="float32"))
        )
        self.unpool_mat.unsqueeze(0)

    def forward(self, x):
        input_shape = list(x.shape)
        # unsqueeze is expand_dims equivalent
        # permute is transpose equivalent
        # view is reshape equivalent
        x = x.unsqueeze(-1)  # bchwx1
        mat = self.unpool_mat.unsqueeze(0)  # 1xshxsw
        ret = torch.tensordot(x, mat, dims=1)  # bxcxhxwxshxsw
        ret = ret.permute(0, 1, 2, 4, 3, 5)
        ret = ret.reshape((-1, input_shape[1], input_shape[2] * 2, input_shape[3] * 2))
        return ret




"""
Module with utility functions for postprocessing.

Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
from typing import Callable, Tuple, List
from .nearest import get_N_closest_pairs_dists, get_N_closest_pairs_idx


Point = Tuple[float, float]
Contour = List[Point]
Cell = Tuple[int, int, Contour]  # id, class


def create_comparator(threshold: float, num_frontier: int) -> Callable[[Contour, Contour], bool]:
    """
    Returns a comparator function between two contours.

    :param threshold: The maximum distance allowed between the num_frontier closest pairs of points to consider two contours equal.
    :type threshold: float
    :param num_frontier: The number of closest pairs of points to consider when comparing contours.
    :type num_frontier: int
    :return: A function that takes two contours as input and returns True if the two contours are equal, False otherwise.
    :rtype: Callable

    This function returns a comparator function between two contours.
    The comparator function takes two contours as input and returns True if the two contours are equal and False otherwise.
    Two contours are considered equal if their num_frontier closest pairs of points are at a distance lower than the specified threshold.
    The function returns the comparator function as a callable object.
    """
    def is_equal(a: Contour, b: Contour) -> bool:
        """
        Two contours are equal if their num_frontier closest pairs
        are at a distance lower than threshold.
        """
        cpairs_dists = get_N_closest_pairs_dists(a, b, num_frontier)
        for dist in cpairs_dists:
            if dist > threshold:
                return False
        return True
    return is_equal


def get_greatest_connected_component(idx: List[int], max_idx: int) -> Tuple[int, int]:
    """
    Given a list of indices, returns the left and right value of the greatest connected component.

    :param idx: A list of indices.
    :type idx: list[int]
    :param max_idx: The maximum index value.
    :type max_idx: int
    :return: A tuple containing the left and right values of the greatest connected component.
    :rtype: tuple(int, int)

    This function takes a list of indices as input and returns the left and right values of the greatest connected component.
    Two indices are considered connected if they differ by one unit.
    The function loops over the indices, keeping track of the left and right indices of the current component.
    When a gap between indices is found, the function checks if the current component is greater than the previous greatest component and updates the left and right indices if it is.
    If the list of indices is circular (i.e., the first and last indices are connected), the function handles this case separately.
    The function returns a tuple containing the left and right values of the greatest connected component.
    """
    idx.sort()
    l_final, l_aux, r_final, r_aux = 0, 0, 0, 0
    found_comp = False
    for k in range(1, len(idx)):
        if idx[k - 1] + 1 != idx[k]:
            found_comp = True
            # End of component
            r_aux = k - 1
            if r_aux - l_aux > r_final - l_final:
                # Save greatest component
                r_final, l_final = r_aux, l_aux
            # Restart left auxiliary index
            l_aux = k
    if not found_comp:
        return idx[0], idx[-1]
    if idx[0] == 0 and idx[-1] == max_idx:
        # Circular case
        r_aux = 1
        while r_aux < len(idx) and idx[r_aux - 1] + 1 == idx[r_aux]:
            r_aux += 1
        r_aux -= 1
        l_aux = len(idx) - 1
        while l_aux >= 0 and idx[l_aux - 1] + 1 == idx[l_aux]:
            l_aux -= 1
        if (r_aux - 0) + (len(idx) - l_aux) > r_final - l_final:
            r_final, l_final = r_aux, l_aux
    return idx[l_final], idx[r_final]


def remove_idx(a: Cell, a_idx: List[int]) -> Cell:
    """
    Removes all the indices in a's contour that are included in a_idx's greatest connected component.

    :param a: The cell containing the contour to modify.
    :type a: Cell
    :param a_idx: A list of indices.
    :type a_idx: list[int]
    :return: The modified cell containing the contour with the specified indices removed.
    :rtype: Cell

    This function takes a cell `a` and a list of indices `a_idx` as input and removes from the contour in `a` all indices that are included in the greatest connected component of `a_idx`.
    The function calls the `get_greatest_connected_component()` function to determine the left and right indices of the greatest connected component.
    The function then creates a new list of indices that includes all the indices in the contour of `a` that are not in the greatest connected component of `a_idx`.
    The function returns a new cell containing the original cell's first two elements and the modified list of indices as the third element.
    """
    left, right = get_greatest_connected_component(a_idx, len(a[2]) - 1)
    res = []
    if left > right:  # Contour is in third position
        res.extend(a[2][right + 1:left])
    elif left <= right:
        res.extend(a[2][right + 1:])
        res.extend(a[2][0:left])
    return (a[0], a[1], res)


def check_order(a: Cell, b: Cell) -> Cell:
    """
    THIS FUNCTION IS NOT IMPLEMENTED YET.

    This function is intended to check whether joining two cells `a` and `b` creates a cross in the middle.
    """
    return True


def join(a: Cell, b: Cell) -> Cell:
    """
    Joins two cells by appending all the points in `b`'s contour to `a`'s contour.

    :param a: The first cell to join.
    :type a: Cell
    :param b: The second cell to join.
    :type b: Cell
    :return: A new cell containing the joined contours.
    :rtype: Cell
    """
    good_order = check_order(a, b)
    if not good_order:
        b[2].reverse()
    a[2].extend(b[2])
    return a


def merge_cells(a: Cell, b: Cell) -> Cell:
    """
    Merges two cells by removing their 30 closest pairs and connecting the endpoints.

    :param a: The first cell to merge.
    :type a: Cell
    :param b: The second cell to merge.
    :type b: Cell
    :return: A new cell containing the merged contours.
    :rtype: Cell
    """
    a_idx, b_idx = get_N_closest_pairs_idx(a[2], b[2], 30)
    a = remove_idx(a, a_idx)
    b = remove_idx(b, b_idx)
    c = join(a, b)
    return c



from .centroids2png import main_with_args as centroids2png_main
from .geojson2pngcsv import main_with_args as geojson2pngcsv_main
from .hovernet2centroids import main_with_args as hovernet2centroids_main
from .hovernet2geojson import main_with_args as hovernet2geojson_main
from .pngcsv2centroids import main_with_args as pngcsv2centroids_main
from .pngcsv2geojson import main_with_args as pngcsv2geojson_main
from .pngcsv2graph import main_with_args as pngcsv2graph_main
from .graph2centroids import main_with_args as graph2centroids_main
from .centroidspng2csv import main_with_args as centroidspng2csv_main
from .png2graph import main_with_args as png2graph_main



"""
Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import numpy as np
import pandas as pd
import os
from sklearn.metrics import confusion_matrix

TEST_DIR = os.path.dirname(os.path.abspath(__file__))

def save(A, B, confusion, name):
    pd.DataFrame(A, columns=['X','Y','class']).to_csv(TEST_DIR + '/centroids/'+ name +'.A.csv', index=False)
    pd.DataFrame(B, columns=['X','Y','class']).to_csv(TEST_DIR + '/centroids/'+ name +'.B.csv', index=False)
    pd.DataFrame(confusion).to_csv(TEST_DIR + '/centroids/'+ name +'.result.csv', index=False, header=False)

def polygon(n, theta=0.01, radius=1):
    paramA = np.linspace(0, 2*np.pi, n+1)
    paramB = paramA + theta
    X, Y = radius*np.cos(paramA), radius*np.sin(paramA)
    A = np.vstack([X, Y, np.ones(n+1)]).transpose()
    X_, Y_ = radius*np.cos(paramB), radius*np.sin(paramB)
    B = np.vstack([X_, Y_, np.ones(n+1)]).transpose()
    return A[:-1], B[:-1]


if __name__=='__main__':
    A = np.random.uniform(low=0, high=1024, size=(100,2))
    A = np.hstack([A, np.ones((100,1))])
    B = A.copy()
    save(A, B, np.array([[100,0],[0,0]]), 'random1')
    B[:,2] = 2
    save(A, B, np.array([[0,100],[0,0]]), 'random2')
    B[:30,2] = 1
    save(A, B, np.array([[30,70],[0,0]]), 'random3')
    A[30:50,2] = 2
    save(A, B, np.array([[30,50],[0,20]]), 'random4')

    A, B = polygon(8)
    save(A, B, np.array([[8,0],[0,0]]), 'circle1')
    A, B = polygon(9)
    save(A, B, np.array([[9,0],[0,0]]), 'circle2')
    A, B = polygon(20)
    save(A, B, np.array([[20,0],[0,0]]), 'circle3')
    A[:10,2] = 2
    B[:10,2] = 2
    save(A, B, np.array([[10,0],[0,10]]), 'circle4')
    A[:5,2] = 1
    save(A, B, np.array([[10,5],[0,5]]), 'circle5')
    B[15:,2] = 2
    save(A, B, np.array([[5,10],[0,5]]), 'circle6')

    A, B = polygon(100, theta=0)
    A[:,:2] *= 8
    B[:,:2] *= 10
    save(A, B, np.array([[100,0],[0,0]]), 'circle7')
    A, B = polygon(50, theta=0)
    A[:,:2] *= 2
    B[:,:2] *= 5
    save(A, B, np.array([[50,0],[0,0]]), 'circle8')
    B[:,2] = 2
    save(A, B, np.array([[0,50],[0,0]]), 'circle9')

    for k in range(10):
        A[:,2] = np.random.randint(1,3,50)
        B[:,2] = np.random.randint(1,3,50)
        conf = confusion_matrix(A[:,2], B[:,2])
        save(A, B, conf, 'random_labels'+str(k+1))

    A, B = polygon(179, theta=2*np.pi/360, radius=20)
    for k in range(10):
        A[:,2] = np.random.randint(1,3,179)
        B[:,2] = np.random.randint(1,3,179)
        conf = confusion_matrix(A[:,2], B[:,2])
        save(A, B, conf, 'random_labels_rotated'+str(k+1))

    N = 3000
    A, B = polygon(N-1, theta=np.pi/(N), radius=200)
    for k in range(30):
        A[:,2] = np.random.randint(1,3,N-1)
        B[:,2] = np.random.randint(1,3,N-1)
        conf = confusion_matrix(A[:,2], B[:,2])
        save(A, B, conf, 'random_labels_rotated_vis'+str(k+1))

    N = 3000
    A, B = polygon(N-1, theta=np.pi/(N), radius=200)
    M = 300
    B = np.vstack([B, np.random.randint(-5, 5, size=(M,3))])
    for k in range(30):
        A[:,2] = np.random.randint(1,3,N-1)
        B[:N-1,2] = np.random.randint(1,3,N-1)
        B[N-1:,2] = np.random.randint(1,3,M)
        conf = confusion_matrix(A[:,2], B[:N-1,2])
        save(A, B, conf, 'extra_points'+str(k+1))



"""
Converts QuPath geojson format into our png <-> csv format.

Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
from typing import Dict, Any, Tuple, List, Optional
import argparse
from skimage.draw import polygon
import numpy as np
import pandas as pd
from tqdm import tqdm
from ..utils.preprocessing import parse_path, create_dir, get_names, save_pngcsv, read_gson


def geojson2pngcsv(gson: List[Dict[str, Any]], num_classes: Optional[int] = 2) -> Tuple[np.ndarray, pd.DataFrame]:
    """
    Computes PNG and CSV labels from GeoJSON.

    :param gson: A list of GeoJSON features.
    :type gson: List[Dict[str, Any]]
    :param num_classes: The number of classes to use in the output CSV. If not provided, it defaults to 2 (tumour and non-tumour).
    :type num_classes: Optional[int]
    :return: A tuple containing the PNG image array and the Pandas DataFrame of the CSV file.
    :rtype: Tuple[np.ndarray, pd.DataFrame]

    This function takes a list of GeoJSON features and generates PNG and CSV labels from them.
    The width and height of the PNG image are assumed to be 1024.
    The function expects the GeoJSON to have specific format, with the geometry stored as a list of coordinates.
    If a feature has a label that is not "tumour" or "non-tumour", the label will be replaced with "ClassN", where N is the class number.
    """
    png = np.zeros((1024, 1024), dtype=np.uint16)
    csv = pd.DataFrame([], columns=['id', 'label'])
    label_parser = {'tumour': 2, 'non-tumour': 1}
    if num_classes != 2:
        label_parser = {"Class" + str(k): k for k in range(1, num_classes + 1)}
    for k, feature in enumerate(gson):
        # Draw filled contour
        contour = feature['geometry']['coordinates'][0]
        if len(contour) <= 1:
            continue
        assert (contour[0] == contour[-1])
        poly = np.array(contour[:-1])
        rr, cc = polygon(poly[:, 0], poly[:, 1], png.shape)
        png[cc, rr] = k + 1

        # Save row to csv
        label = feature['properties']['classification']['name']
        label = label_parser[label]
        csv = pd.concat([csv, pd.DataFrame([[k + 1, label]], columns=['id', 'label'])])
    return png, csv


def _create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--gson-dir', type=str, required=True,
                        help='Path to the geojson files.')
    parser.add_argument('--png-dir', type=str, required=True,
                        help='Path to save the png files.')
    parser.add_argument('--csv-dir', type=str, required=True,
                        help='Path to save the csv files.')
    parser.add_argument('--num-classes', type=int, default=2)
    return parser


def main_with_args(args):
    gson_dir = parse_path(args.gson_dir)
    png_dir = parse_path(args.png_dir)
    csv_dir = parse_path(args.csv_dir)

    create_dir(png_dir)
    create_dir(csv_dir)

    names = get_names(gson_dir, '.geojson')
    for name in tqdm(names):
        gson = read_gson(name, gson_dir)
        png, csv = geojson2pngcsv(gson, args.num_classes)
        save_pngcsv(png, csv, png_dir, csv_dir, name)


def main():
    parser = _create_parser()
    args = parser.parse_args()
    main_with_args(args)



"""
Merges .centroids.csv and .GT_cells.png into .class.csv files.

Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
from typing import List, Tuple
import argparse
from argparse import Namespace
from ..utils.preprocessing import get_names, parse_path, create_dir, read_png, save_csv, read_centroids, get_centroid_by_id
import pandas as pd
import numpy as np
from ..utils.nearest import generate_tree, find_nearest
from tqdm import tqdm


def extract_centroids(img: np.ndarray) -> List[Tuple[int, int, int]]:
    """
    Extracts the centroids of cells from a labeled image. The third coordinates is the index.

    :param img: A 2D NumPy array representing the labeled image. Each unique non-zero value represents a different cell.
    :type img: np.ndarray
    :return: A list of tuples containing the x and y coordinates of the centroid, and the cell index.
    :rtype: List[Tuple[int,int,int]]
    """
    centroids = []
    for idx in np.unique(img):
        if idx != 0:
            x, y = get_centroid_by_id(img, idx)
            if x == -1:
                continue
            centroids.append((x, y, idx))
    return centroids


def centroidspng2csv(centroids_file: np.ndarray, png_file: np.ndarray) -> pd.DataFrame:
    """
    Converts a PNG file with cell labels and a CSV file with cell centroids into a CSV file
    associating each cell label with the closest centroid.

    :param centroids_file: A NumPy array with three columns, representing the X and Y coordinates
                           and class of each centroid.
    :type centroids_file: np.ndarray
    :param png_file: A NumPy array with the same dimensions as the corresponding image file,
                     where each pixel is labeled with an integer representing the cell it belongs to.
    :type png_file: np.ndarray
    :return: A Pandas DataFrame with two columns: the first represents the cell ID from the PNG file,
             and the second represents the class of the closest centroid from the CSV file.
    :rtype: pd.DataFrame

    This function first generates a k-d tree from the centroids in the `centroids_file` array.
    For each centroid in the `png_file` array, the function finds the closest centroid in the
    `centroids_file` array using the k-d tree. It then associates the cell label from the PNG file
    with the class of the closest centroid in a list.
    Finally, the function returns a Pandas DataFrame containing the list of cell IDs and centroid
    classes.
    """
    centroids_tree = generate_tree(centroids_file)
    png_centroids = extract_centroids(png_file)
    csv_list = []
    for point in png_centroids:
        closest_id = find_nearest(point, centroids_tree)
        closest = centroids_file[closest_id]
        csv_list.append([point[2], closest[2]])
    return pd.DataFrame(csv_list)


def _create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--centroids-dir', type=str, required=True, help='Path to folder containing .centroids.csv.')
    parser.add_argument('--png-dir', type=str, required=True, help='Path to folder containing .GT_cells.png.')
    parser.add_argument('--csv-dir', type=str, required=True, help='Path to folder where to save .class.csv.')
    return parser


def main_with_args(args: Namespace) -> None:
    centroids_dir = parse_path(args.centroids_dir)
    png_dir = parse_path(args.png_dir)
    csv_dir = parse_path(args.csv_dir)
    create_dir(csv_dir)
    names_centroids = sorted(get_names(centroids_dir, '.centroids.csv'))
    names_png = sorted(get_names(png_dir, '.GT_cells.png'))
    names = list(set(names_png).intersection(set(names_centroids)))
    for name in tqdm(names):
        centroids_file = read_centroids(name, centroids_dir)
        png_file = read_png(name, png_dir)
        csv_file = centroidspng2csv(centroids_file, png_file)
        save_csv(csv_file, csv_dir, name)
    return


def main():
    parser = _create_parser()
    args = parser.parse_args()
    main_with_args(args)



import inspect
import logging
import os
import shutil
import cv2
import numpy as np
from scipy import ndimage


####
def normalize(mask, dtype=np.uint8):
    return (255 * mask / np.amax(mask)).astype(dtype)


####
def get_bounding_box(img):
    """Get bounding box coordinate information."""
    rows = np.any(img, axis=1)
    cols = np.any(img, axis=0)
    rmin, rmax = np.where(rows)[0][[0, -1]]
    cmin, cmax = np.where(cols)[0][[0, -1]]
    # due to python indexing, need to add 1 to max
    # else accessing will be 1px in the box, not out
    rmax += 1
    cmax += 1
    return [rmin, rmax, cmin, cmax]


####
def cropping_center(x, crop_shape, batch=False):
    """Crop an input image at the centre.

    Args:
        x: input array
        crop_shape: dimensions of cropped array
    
    Returns:
        x: cropped array
    
    """
    orig_shape = x.shape
    if not batch:
        h0 = int((orig_shape[0] - crop_shape[0]) * 0.5)
        w0 = int((orig_shape[1] - crop_shape[1]) * 0.5)
        x = x[h0 : h0 + crop_shape[0], w0 : w0 + crop_shape[1]]
    else:
        h0 = int((orig_shape[1] - crop_shape[0]) * 0.5)
        w0 = int((orig_shape[2] - crop_shape[1]) * 0.5)
        x = x[:, h0 : h0 + crop_shape[0], w0 : w0 + crop_shape[1]]
    return x


####
def rm_n_mkdir(dir_path):
    """Remove and make directory."""
    if os.path.isdir(dir_path):
        shutil.rmtree(dir_path)
    os.makedirs(dir_path)


####
def mkdir(dir_path):
    """Make directory."""
    if not os.path.isdir(dir_path):
        os.makedirs(dir_path)


####
def get_inst_centroid(inst_map):
    """Get instance centroids given an input instance map.

    Args:
        inst_map: input instance map
    
    Returns:
        array of centroids
    
    """
    inst_centroid_list = []
    inst_id_list = list(np.unique(inst_map))
    for inst_id in inst_id_list[1:]:  # avoid 0 i.e background
        mask = np.array(inst_map == inst_id, np.uint8)
        inst_moment = cv2.moments(mask)
        inst_centroid = [
            (inst_moment["m10"] / inst_moment["m00"]),
            (inst_moment["m01"] / inst_moment["m00"]),
        ]
        inst_centroid_list.append(inst_centroid)
    return np.array(inst_centroid_list)


####
def center_pad_to_shape(img, size, cval=255):
    """Pad input image."""
    # rounding down, add 1
    pad_h = size[0] - img.shape[0]
    pad_w = size[1] - img.shape[1]
    pad_h = (pad_h // 2, pad_h - pad_h // 2)
    pad_w = (pad_w // 2, pad_w - pad_w // 2)
    if len(img.shape) == 2:
        pad_shape = (pad_h, pad_w)
    else:
        pad_shape = (pad_h, pad_w, (0, 0))
    img = np.pad(img, pad_shape, "constant", constant_values=cval)
    return img


####
def color_deconvolution(rgb, stain_mat):
    """Apply colour deconvolution."""
    log255 = np.log(255)  # to base 10, not base e
    rgb_float = rgb.astype(np.float64)
    log_rgb = -((255.0 * np.log((rgb_float + 1) / 255.0)) / log255)
    output = np.exp(-(log_rgb @ stain_mat - 255.0) * log255 / 255.0)
    output[output > 255] = 255
    output = np.floor(output + 0.5).astype("uint8")
    return output


####
def log_debug(msg):
    frame, filename, line_number, function_name, lines, index = inspect.getouterframes(
        inspect.currentframe()
    )[1]
    line = lines[0]
    indentation_level = line.find(line.lstrip())
    logging.debug("{i} {m}".format(i="." * indentation_level, m=msg))


####
def log_info(msg):
    frame, filename, line_number, function_name, lines, index = inspect.getouterframes(
        inspect.currentframe()
    )[1]
    line = lines[0]
    indentation_level = line.find(line.lstrip())
    logging.info("{i} {m}".format(i="." * indentation_level, m=msg))


def remove_small_objects(pred, min_size=64, connectivity=1):
    """Remove connected components smaller than the specified size.

    This function is taken from skimage.morphology.remove_small_objects, but the warning
    is removed when a single label is provided. 

    Args:
        pred: input labelled array
        min_size: minimum size of instance in output array
        connectivity: The connectivity defining the neighborhood of a pixel. 
    
    Returns:
        out: output array with instances removed under min_size

    """
    out = pred

    if min_size == 0:  # shortcut for efficiency
        return out

    if out.dtype == bool:
        selem = ndimage.generate_binary_structure(pred.ndim, connectivity)
        ccs = np.zeros_like(pred, dtype=np.int32)
        ndimage.label(pred, selem, output=ccs)
    else:
        ccs = out

    try:
        component_sizes = np.bincount(ccs.ravel())
    except ValueError:
        raise ValueError(
            "Negative value labels are not supported. Try "
            "relabeling the input with `scipy.ndimage.label` or "
            "`skimage.morphology.label`."
        )

    too_small = component_sizes < min_size
    too_small_mask = too_small[ccs]
    out[too_small_mask] = 0

    return out



"""
Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import pytest
import os
import numpy as np
import pandas as pd

from tumourkit.preprocessing.pngcsv2graph import pngcsv2graph
from tumourkit.utils.preprocessing import get_names, parse_path, read_labels

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
PNGCSV_DIR = parse_path(TEST_DIR) + 'pngcsv/'
THRESHOLD = 0.1


@pytest.mark.parametrize("name", get_names(PNGCSV_DIR, '.GT_cells.png')[:5])
def test_indices(name):
    png, csv = read_labels(name, PNGCSV_DIR, PNGCSV_DIR)
    false_img = np.ones((*png.shape, 3)) * png.reshape((*png.shape, 1))
    false_img = false_img.astype(np.uint8)
    graph = pngcsv2graph(false_img, png, csv)
    graph.set_index('id', inplace=True)
    graph.sort_index(inplace=True)
    csv.columns = ['id', 'class']
    csv.set_index('id', inplace=True)
    csv.sort_index(inplace=True)
    join = pd.merge(left=csv, right=graph, 
         left_index=True, right_index=True, how='outer').dropna()
    assert((join.class_x == join.class_y).all())
